#! /usr/bin/env python3
#
# Test(s) for ../main.py
#
# Notes:
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_main.py
# For tips on pytest monkeypatch, see
#   https://stackoverflow.com/questions/38723140/i-want-to-use-stdin-in-a-pytest-test
#

"""Unit tests for main module"""

# Standard packages
from argparse import ArgumentParser
import io
import sys

# Installed packages
import pytest

# Local packages
from mezcla import debug
from mezcla import system
from mezcla import tpo_common as tpo
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla.unittest_wrapper import trap_exception

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.main as THE_MODULE

class MyArgumentParser(ArgumentParser):
    """Version of ArgumentParser that doesn't exit upon failure"""

    def __init__(self, *args, **kwargs):
        """Constructor (relegating all to parent)"""
        super().__init__(*args, **kwargs)

    def exit(self, status=0, message=None):
        """Version of exit that doesn't really exit"""
        if message:
            self._print_message(message, sys.stderr)


class TestMain(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.derive_tested_module_name(__file__)
    ## HACK: globals for use in embedded classes (TODO: move into Test class below)
    input_processed = None
    main_step_invoked = None
    process_line_count = -1

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_script_options(self):
        """Makes sure script option specifications are parsed OK"""
        debug.trace(4, f"in test_script_options(); self={self}")
        class Test(THE_MODULE.Main):
            """"Dummy test class"""
            argument_parser = MyArgumentParser
            ## OLD: skip_args = True
            ## TODO: rename as TestMain?; drop MyArgumentParser?

        # note: verbose-mode and file-path used as main supplies --verbose and filename
        FULL_NAME_OPT = "full-name"
        FULL_NAME_UNDER = FULL_NAME_OPT.replace("-", "_")
        VERBOSE_OPT = "verbose-mode"
        VERBOSE_UNDER = VERBOSE_OPT.replace("-", "_")
        FILE_PATH_OPT = "file-path"
        FILE_PATH_UNDER = FILE_PATH_OPT.replace("-", "_")
        JOHN_DOE = "John Doe"
        MY_FILE = "my-file.txt"
        # note: format is ("option", "description", "default"), or just "option"
        app = Test(text_options=[(FULL_NAME_OPT, "Full name", JOHN_DOE)],
                   boolean_options=[(VERBOSE_OPT, "testing verbose option", True)],
                   positional_options=[(FILE_PATH_OPT, "Path for input")],
                   runtime_args=[f"--{VERBOSE_OPT}", MY_FILE],
                   )
        #
        debug.trace_expr(5, app.parsed_args)
        # note: underscores used in argparse hash keys
        self.do_assert(app.parsed_args.get(FULL_NAME_UNDER) == JOHN_DOE)
        self.do_assert(not app.parsed_args.get(FULL_NAME_OPT))
        self.do_assert(app.parsed_args.get(VERBOSE_UNDER))
        self.do_assert(not app.parsed_args.get(VERBOSE_OPT))
        # note: dashes retained in positional args
        self.do_assert(not app.parsed_args.get(FILE_PATH_UNDER))
        self.do_assert(app.parsed_args.get(FILE_PATH_OPT) == MY_FILE)
        #
        # note: dashes converted to underscores by argparse wrappers
        self.do_assert(app.get_parsed_option(FULL_NAME_UNDER, allow_under=True)
                       == JOHN_DOE)
        self.do_assert(app.get_parsed_option(FULL_NAME_OPT) == JOHN_DOE)
        self.do_assert(app.get_parsed_option(VERBOSE_UNDER, allow_under=True))
        self.do_assert(app.get_parsed_option(VERBOSE_OPT))
        self.do_assert(app.get_parsed_argument(FILE_PATH_UNDER, allow_under=True)
                       == MY_FILE)
        self.do_assert(app.get_parsed_argument(FILE_PATH_OPT) == MY_FILE)
        debug.trace(5, "out test_script_options")

    @pytest.mark.xfail                   # TODO: remove xfail
    @trap_exception
    def test_script_without_input(self):
        """Makes sure script class without input doesn't process input and that
        the main step gets invoked"""
        debug.trace(4, f"in test_script_without_input(); self={self}")

        ## OLD
        ## # This avoids flaky stderr due to other tests
        ## tpo.restore_stderr()

        # Create scriptlet checking for input and processing main step
        # TODO: rework with external script as argparse exits upon failure
        class Test(THE_MODULE.Main):
            """"Dummy test class"""
            argument_parser = MyArgumentParser

            def setup(self):
                """Post-argument parsing processing: just displays context"""
                tpo.debug_format("in Test.setup(): self={s}", 5, s=self)
                TestMain.process_line_count = 0
                tpo.trace_current_context(label="Test.setup", 
                                          level=tpo.QUITE_DETAILED)
                tpo.trace_object(self, tpo.QUITE_DETAILED, "Test instance")
            #
            def process_line(self, line):
                """Dummy input processing"""
                tpo.debug_format("in Test.process_line({l}): self={s}", 5,
                                 l=line, s=self)
                TestMain.input_processed = True
                TestMain.process_line_count += 1

            #
            def run_main_step(self):
                """Dummy main step"""
                tpo.debug_format("in Test.run_main_step()): self={s}", 5,
                                 s=self)
                TestMain.main_step_invoked = True

        # Test scriptlet with test script as input w/ and w/o input enabled
        debug.trace(6, "app1")
        TestMain.input_processed = None
        ## TEST: MAIN_SCRIPT = "main.py"         # TOOD3: derive from __file__
        MAIN_SCRIPT = __file__
        app1 = Test(skip_input=False, manual_input=False, runtime_args=[__file__], program=MAIN_SCRIPT)          # avoids pytest if invoked via it
        try:
            app1.run()
        except:
            tpo.print_stderr("Exception during app1.run: {exc}",
                             exc=tpo.to_string(sys.exc_info()))
        self.do_assert(TestMain.input_processed)
        #
        TestMain.input_processed = None
        debug.trace(6, "app2")
        ## NOTE: This produces an extraneous error, but the test class still executes:
        ##   test_main.py: error: unrecognized arguments: tests/test_main.py
        ## BAD:
        app2 = Test(skip_input=True, manual_input=True, runtime_args=[__file__], program=MAIN_SCRIPT)          # avoids pytest if invoked via it
        ## TODO2: app3 = Test(skip_input=True, manual_input=True, runtime_args=["-", "<", __file__], program=MAIN_SCRIPT)        # avoids pytest if invoked via it
        try:
            app2.run()
        except:
            tpo.print_stderr("Exception during app2.run: {exc}",
                             exc=tpo.to_string(sys.exc_info()))
        self.do_assert(not TestMain.input_processed)

        # Test scriptlet w/ input disabled and without arguments
        # note: auto_help disabled so that no arguments needed
        TestMain.main_step_invoked = None
        debug.trace(6, "app3")
        app3 = Test(skip_input=True, manual_input=True, auto_help=False, runtime_args=[])
        try:
            app3.run()
        except:
            tpo.print_stderr("Exception during app3.run: {exc}",
                             exc=tpo.to_string(sys.exc_info()))
        self.do_assert(not TestMain.input_processed)
        self.do_assert(TestMain.main_step_invoked)
        
        ## TODO2: Test scriptlet w/ all input disabled
        ## # note: auto_help disabled so that no arguments needed
        ## debug.trace(6, "app4")
        ## app4 = Test(skip_input=True, manual_input=True, auto_help=False, runtime_args=[])
        ## ...
        debug.trace(5, "out test_script_without_input")

    @pytest.mark.xfail                   # TODO: remove xfail
    @trap_exception
    def test_perl_arg(self):
        """Make sure perl-style arg can be parsed"""
        # TODO: create generic app-creation helper
        debug.trace(4, f"in test_perl_arg(); self={self}")
        class Test(THE_MODULE.Main):
            """"Dummy test class"""
            argument_parser = MyArgumentParser
            ## OLD: skip_args = True

        # Test with and without Perl support
        app = Test(boolean_options=[("fubar", "testing fubar option")],
                   runtime_args=["-fubar"], perl_switch_parsing=True)
        debug.trace_expr(5, app.parsed_args)
        #
        self.do_assert(system.to_bool(app.parsed_args.get("fubar")))
        #
        app = Test(boolean_options=[("fubar", "testing fubar option")],
                   runtime_args=["-fubar"], perl_switch_parsing=False)
        debug.trace_expr(5, app.parsed_args)
        # NOTE: this ensures that is None and not 0
        self.do_assert(app.parsed_args.get("fubar") is None)
        debug.trace(5, "out test_perl_arg")


class TestMain2:
    """Another class for testcase definition
    Note: Needed to avoid error with pytest due to inheritance with unittest.TestCase via TestWrapper (e.g., capsys)"""

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_input_modes(self, capsys, monkeypatch):
        """Make sure input processed OK with respect to line/para/file mode"""
        debug.trace(4, f"in test_input_modes({capsys}); self={self}")
        contents = "1\n\n2\n\n\n3\n\n\n\n4\n\n\n\n"
        num_lines = len(contents.split("\n"))
        pre_captured = capsys.readouterr()
        # line input
        monkeypatch.setattr('sys.stdin', io.StringIO(contents))
        main = THE_MODULE.Main(skip_args=True, auto_help=False)
        main.run()
        captured = capsys.readouterr()
        ## TODO: assert(TestMain.process_line_count == num_lines)
        assert(num_lines == len(captured.out.split("\n")))
        debug.trace_expr(5, main, num_lines)
        # paragraph input
        monkeypatch.setattr('sys.stdin', io.StringIO(contents))
        main = THE_MODULE.Main(paragraph_mode=True, skip_args=True, auto_help=False)
        main.run()
        captured = capsys.readouterr()
        ## TODO: assert(TestMain.process_line_count == 4)
        assert(num_lines == len(captured.out.split("\n")))
        debug.trace_expr(5, main, num_lines)
        # file input
        monkeypatch.setattr('sys.stdin', io.StringIO(contents))
        main = THE_MODULE.Main(file_input_mode=True, skip_args=True, auto_help=False)
        main.run()
        captured = capsys.readouterr()
        ## TODO: assert(TestMain.process_line_count == 1)
        assert(num_lines == len(captured.out.split("\n")))
        debug.trace_expr(5, main, num_lines, pre_captured)
        debug.trace(5, "out test_input_modes")

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_missing_newline(self, capsys, monkeypatch):
        """Make sure file with missing newline at end processed OK"""
        debug.trace(4, f"in test_missing_newline({capsys}); self={self}")
        contents = "1\n2\n3"
        num_lines = len(contents.split("\n"))
        _pre_captured = capsys.readouterr()
        monkeypatch.setattr('sys.stdin', io.StringIO(contents))
        main = THE_MODULE.Main(skip_args=True, auto_help=False)
        main.run()
        captured = capsys.readouterr()
        # note: 1 extra line ('') and 1 extra character (final newline)
        assert((1 + num_lines) == len(captured.out.split("\n")))
        assert((1 + len(contents)) == len(captured.out))
        ## TODO: self.do_assert(TestMain.process_line_count == 3)
        debug.trace_expr(5, main, num_lines)
        debug.trace(5, "out test_missing_newline")

    def test_has_parsed_option_hack(self):
        """Make sure (temporarily hacked) has_parsed_option differs from has_parsed_option_old"""
        debug.trace(4, f"in test_has_parsed_option_hack(); self={self}")
        ok_arg = "ok"
        missing_arg = "missing"
        main = THE_MODULE.Main(skip_args=True, auto_help=False)
        main.parsed_args = {ok_arg: True}
        assert(main.has_parsed_option_old(ok_arg) == main.has_parsed_option(ok_arg))
        assert(main.has_parsed_option_old(missing_arg) != main.has_parsed_option(missing_arg))
        debug.trace(5, "out test_has_parsed_option_hack")

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# This currently includes adhoc code for modifying how the tests are run.
# For example, there is an attempt at getting pytest not to swallow exceptions.
# In addition, there is code for refined results reporting with thresholding.
# Features are disabled by default so that pytest runs normally.
#

"""Adhoc configuration for pytest: disabled by default"""

# Standard packages
## OLD: import os

# Installed packages
import pytest

# Local packages
from mezcla import debug
from mezcla import system

# Environment options
ENABLE_PYTEST_RAISE = system.getenv_bool(
    "ENABLE_PYTEST_RAISE", False,
    desc="Raise exceptions in test code")
ENABLE_PYTEST_REPORTING = system.getenv_bool(
    "ENABLE_PYTEST_REPORTING", False,
    desc="Apply hooks for pytest threshold checks")

#--------------------------------------------------------------------------------

# Add optional hooks to disable exception swallowing
if ENABLE_PYTEST_RAISE:
    # Note:
    # - Better to Use @trap_exception: See unittest_wrapper.py
    #   and tests/template.py. In addition, using --capture=none can
    #   be helpful and likewise --runxfail for tests marked xfail.
    # - Based on https://stackoverflow.com/questions/62419998/how-can-i-get-pytest-to-not-catch-exceptions.
    # - Also see https://docs.pytest.org/en/7.1.x/reference/reference.html.

    def pytest_exception_interact(node, call, report):
        """Called when an exception was raised which can potentially be interactively handled"""
        debug.trace(5, f"pytest_exception_interact{((node, call, report))}")
        raise call.excinfo.value

    def pytest_internalerror(excrepr, excinfo):
        """Return True to suppress the fallback handling of printing an INTERNALERROR message directly to sys.stderr"""
        debug.trace(5, f"pytest_internalerror{(excrepr, excinfo)}")
        raise excinfo.value

#--------------------------------------------------------------------------------
# Note: this is experimental code that was pre-maturely commited

# Add optional hooks to allow for better test reporting along with thresholds.
if ENABLE_PYTEST_REPORTING:
    # TODO2:
    # - Document an overview of the processing (e.g., high-level intention)
    # - Explain how to invoke.
    # - Provide pointers to similar code.
    # - Don't commit such new code without warning!
    
    # Dictionary to store test results
    test_results = {
        'total': 0,
        'passed': 0,
        'failed': 0,
        'skipped': 0,
        'xfailed': 0,
        'xpassed': 0
    }

    @pytest.hookimpl(tryfirst=True)
    def pytest_runtest_setup(item):
        """Called before each test is run. Ensure total count is unique per test."""
        if item.nodeid not in test_results:
            test_results['total'] += 1
            test_results[item.nodeid] = True
    
    @pytest.hookimpl(tryfirst=True)
    def pytest_runtest_makereport(item, call):
        """Called to create a test report for each test phase."""
        report = pytest.TestReport.from_item_and_call(item, call)
    
        # Count passed, failed, and skipped tests
        if report.when == "call":
            if report.outcome == "passed":
                test_results['passed'] += 1
            elif report.outcome == "skipped":
                test_results['skipped'] += 1
            elif report.outcome == "failed":
                test_results['failed'] += 1
    
        if report.outcome == "skipped" and call.excinfo is None:
            test_results['skipped'] += 1
    
    def pytest_addoption(parser):
        """Add custom options to pytest command-line."""
        ## TODO4: Use getenv_float for default to simplify usgae
        parser.addoption(
            "--pass-threshold",
            action="store",
            default="80",
            help="Pass threshold percentage for tests",
        )
    
    def helper_calculate_pass_percentage():
        """Calculate the pass percentage for the test session."""
        actual_total = test_results['total'] - test_results['skipped']
        if actual_total == 0:
            return 0.0
        return (test_results['passed'] / actual_total) * 100
    
    @pytest.hookimpl(tryfirst=True)
    def pytest_sessionfinish(session):
        """Pytest hook: Called after the test session finishes."""
        pass_threshold = int(session.config.getoption("--pass-threshold"))
        pass_percentage = helper_calculate_pass_percentage()
    
        # Print test summary
        print("\n\nTest Summary\n" + "=" * 15)
        print(f"Total Tests: {test_results['total']}")
        print(f"Passed: {test_results['passed']}")
        print(f"Failed: {test_results['failed']}")
        print(f"Skipped: {test_results['skipped']}")
        print(f"Pass Percentage: {pass_percentage:.2f}% (Threshold: {pass_threshold}%)")
    
        # Set exit status based on pass percentage
        if pass_percentage < pass_threshold:
            print("[FAILED]")
            session.exitstatus = pytest.ExitCode.TESTS_FAILED
        else:
            session.exitstatus = pytest.ExitCode.OK
            print("[PASSED]")

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    system.print_stderr("Error: Not intended to be invoked directly")



#! /usr/bin/env python3
#
# Test(s) for ../run_albert_classifier.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_run_albert_classifier.py
#

"""Tests for run_albert_classifier module"""

# Standard packages
## NOTE: this is empty for now

# Installed packages
import pytest

# Local packages
from mezcla import debug
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla import system

# Note: Two references are used for the module to be tested:
#    THE_MODULE:               module object (e.g., <module 'mezcla.main' ...>)
#    TestIt.script_module:     dotted module path (e.g., "mezcla.main")
try:
    import mezcla.run_albert_classifier as THE_MODULE
    import mezcla.bert_multi_classification as THE_MODULE
except:
    THE_MODULE = None
    system.print_exception_info("THE_MODULE import")

class TestRunAlbertClassifier(TestWrapper):
    """Class for testcase definition"""
    # note: script_module used in argument parsing sanity check (e.g., --help)
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    @pytest.mark.xfail                   # TODO: remove xfail
    ## DEBUG: @trap_exception            # TODO: remove when debugged
    def test_01_data_file(self):
        """Tests run_script w/ data file"""
        # Warning: see notes above about potential issues with run_script-based tests.
        debug.trace(4, f"TestIt.test_01_data_file(); self={self}")
        data = ["TODO1", "TODO2"]
        system.write_lines(self.temp_file, data)
        ## TODO: add use_stdin=True to following if no file argument
        output = self.run_script(options="", data_file=self.temp_file)
        self.do_assert(my_re.search(r"TODO-pattern", output.strip()))
        return

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Test(s) for ../os_utils.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_os_utils.py
#
# TODO3: apply mezcla conventions (see tests/template.py).
#

"""Tests for os_utils module"""

# Standard packages
import types

# Installed packages
import pytest

# Local packages
from mezcla import debug
from mezcla import os_utils
from mezcla import system
from mezcla.unittest_wrapper import TestWrapper, invoke_tests

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	         global module object
#    TestWrapper.script_module:  path to file
THE_MODULE = os_utils

class TestOsUtils(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    @classmethod
    def setupClass(cls):
        """Per-class initialization"""  # why not docstring inherited (or pylint bug?)
        debug.trace(5, f"TestOsUtils.setupClass({cls})")
        debug.trace_expr(5, __file__, THE_MODULE)
        super().setUpClass(filename=__file__, module=THE_MODULE)
    
    def test_split_extension(self):
        """Ensure test_split_extension works as expected"""
        assert(os_utils.split_extension("fubar.txt") == ("fubar", ".txt"))
        assert(os_utils.split_extension("abc.def.txt") == ("abc.def", ".txt"))

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_02_no_other_functions(self):
        """Make sure no other functions defined in the module"""
        ## TODO3: put helper for this in tests/common_module.py; also,
        ## use monkeypatch to check for new function
        debug.trace(4, f"TestIt.test_02_no_other_functions(); self={self}")
        defined_functions = [
            m for m in dir(THE_MODULE)
            if isinstance(getattr(THE_MODULE, m), types.FunctionType)]
        self.do_assert(not (system.difference(
            defined_functions, ['main', 'split_extension'])))
        return


if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Downloads transcript for YouTube video
#
# Based on following:
#   https://github.com/jdepoix/youtube-transcript-api/issues/234 [expose functionality for formatting timestamps in textual formats]
#

"""
Download YouTube transcript

Sample usages:
   {script} 'https://www.youtube.com/watch?v=1KcdgFxmnb4' > Caravaggio-examples.txt

   id=3UWxmt7VAlU; {script} "$id" > edward-loper-doc.txt
"""

# Standard modules
## TODO: import json

# Intalled module
import youtube_transcript_api as ytt_api
from youtube_transcript_api import formatters

# Local modules
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import Main
from mezcla import misc_utils
from mezcla.my_regex import my_re
from mezcla import system

# Constants
TL = debug.TL

#-------------------------------------------------------------------------------
# Note: custom class due lack of help by youtube_transcript_api developers

class YouTubeLikeFormatter(formatters._TextBasedFormatter):      # pylint: disable=protected-access
    """Uses format similar to that under YouTube's Transcript pane:
         0:16 This is the city after a storm, ...
         0:23 "Once you learn to see as an artist, ..."""
    
    def _format_timestamp(self, hours, mins, secs, _ms):
        # format as HH:MM:SS w/ 00 hour omitted and with leading zeros dropped
        timestamp = "{:02d}:{:02d}:{:02d}".format(hours, mins, secs)
        timestamp = my_re.sub("^00:", "", timestamp)
        timestamp = my_re.sub("^0", "", timestamp)
        return timestamp

    def _format_transcript_header(self, lines):
        return "Transcript\n\n" + "\n".join(lines) + "\n"

    def _format_transcript_helper(self, i, time_text, line):
        # drops second timestamp (e.g., "00:00:28.500 --> 00:00:30.060" => "00:00:28.500")
        time_text = my_re.sub(r" --> \S+", "", time_text)
        return "{} {}".format(time_text, line['text'])

#-------------------------------------------------------------------------------

def main():
    """Entry point"""
    debug.trace(TL.USUAL, f"main(): script={system.real_path(__file__)}")

    # Parse command line options, show usage if --help given
    main_app = Main(description=__doc__.format(script=gh.basename(__file__)),
                    skip_input=False, manual_input=False)
    debug.assertion(main_app.parsed_args)
    url = main_app.filename
    video_id= url
    YOUTUBE_PREFIX = "https://www.youtube.com/watch?v="
    if url.startswith(YOUTUBE_PREFIX):
        video_id = url[len(YOUTUBE_PREFIX):]
    else:
        debug.assertion("youtube.com" not in url)
        url = YOUTUBE_PREFIX + url
    ## TODO: video_id = my_re.sub(r"(https://)?www.youtube.com/watch\?v=", "", main_app.filename)
    debug.trace_expr(5, url, video_id)

    # Download the transcript and print using YouTubeLike-format
    # note: youtube_transcript_api
    print(misc_utils.get_formatted_date())
    print("")
    print(url)
    print("")
    try:
        transcript = ytt_api.YouTubeTranscriptApi.get_transcript(video_id)
        print(YouTubeLikeFormatter().format_transcript(transcript))
    except:
        system.print_exception_info("transcript access")
        print("n/a")

    return

#-------------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context(level=TL.QUITE_VERBOSE)
    main()



#! /usr/bin/env python3
#
# Wrapper class around unittest.TestCase
#
# Notes:
# - Based on template.py used in older test scripts
# - Creates per-test temp file, based on same class-wide temp-base file.
# - To treat temp-base as a subdirectory, set use_temp_base_dir to True in 
#   class member initialiation section.
# - Changes to temporary directory/file should be synchronized with ../main.py.
# - Overriding the temporary directory can be handy during debugging; however,
#   you might need to specify different ones if you invoke helper scripts. See
#   tests/test_extract_company_info.py for an example.
# - It can be confusing debugging tests that use run_script, because the trace level
#   is raised by default. To disable this, set the SUB_DEBUG_LEVEL as follows:
#      l=5; DEBUG_LEVEL=$l SUB_DEBUG_LEVEL=$l pytest -s tests/test_spell.py
#   See glue_helper.py for implementation along with related ALLOW_SUBCOMMAND_TRACING.
# TODO:
# - * Clarify TEMP_BASE vs. TEMP_FILE usage.
#   via glue_helpers.py: default base prefix vs fixed override
# - Add TEMP_DIR for more direct specification.
# - Clarify that this can co-exist with pytest-based tests (see tests/test_main.py).
# TODO2:
# - Clean up script_file usage (and unncessary settings in test scripts).
#
#-------------------------------------------------------------------------------
# Sample test (streamlined version of test_simple_main_example.py):
#
#   import unittest
#   from mezcla import system
#   from mezcla.unittest_wrapper import TestWrapper, invoke_tests
#
#   class TestIt(TestWrapper):
#       """Class for testcase definition"""
#       script_module = TestWrapper.derive_tested_module_name(__file__)
#  
#       def test_simple_data(self):
#           """Make sure simple data sample processed OK"""
#           system.write_file(self.temp_file, "really fubar")
#           output = self.run_script("--check-fubar", self.temp_file)
#           self.assertTrue("really" in output)
#
#   if __name__ == '__main__':
#      invoke_tests(__file__)
#-------------------------------------------------------------------------------
# TODO:
# - Add method to invoke unittest.main(), so clients don't need to import unittest.
#

"""Unit test support class"""

# Standard packages
import inspect
import os
import sys
import tempfile
import unittest
from typing import (
    Optional, Callable, Any, Tuple,
)
## DEBUG: sys.stderr.write(f"{__file__=}\n")

# Installed packages
import pytest

# Local packages
# note: Disables TEMP_FILE default used by glue_helpers.py.
PRESERVE_TEMP_FILE_LABEL = "PRESERVE_TEMP_FILE"
if PRESERVE_TEMP_FILE_LABEL not in os.environ:
    ## DEBUG: sys.stderr.write(f"Setting {PRESERVE_TEMP_FILE_LABEL}\n")
    os.environ[PRESERVE_TEMP_FILE_LABEL] = "1"
import mezcla
from mezcla import debug
from mezcla import glue_helpers as gh
## BAD: from mezcla.main import DISABLE_RECURSIVE_DELETE
DISABLE_RECURSIVE_DELETE = gh.DISABLE_RECURSIVE_DELETE
from mezcla.misc_utils import string_diff
from mezcla.my_regex import my_re
from mezcla import system
## DEBUG: debug.trace_expr(6, __file__)
from mezcla.debug import IntOrTraceLevel

# Constants (e.g., environment options)

TL = debug.TL
KEEP_TEMP = gh.KEEP_TEMP
PRUNE_TEMP = system.getenv_bool(
    "PRUNE_TEMP", False,
    desc="Delete temporary files ahead of time")
TODO_FILE = "TODO FILE"
TODO_MODULE = "TODO MODULE"
TEMP_BASE = gh.TEMP_BASE

# Note: the following is for transparent resolution of dotted module names
# for invocation of scripts via 'python -m package.module'. This is in support
# of transitioning from the old way of importing packages via 'import module'
# instead of 'import package.module'. (The former required that package be
# explicitly specified in the python path, such as via 'PYTHONPATH=package-dir:...'.)
THIS_PACKAGE = getattr(mezcla.debug, "__package__", None)
debug.assertion(THIS_PACKAGE == "mezcla")

# Environment options
VIA_UNITTEST = system.getenv_bool(
    "VIA_UNITTEST", False,
    description="Run tests via unittest instead of pytest")
PROFILE_CODE = system.getenv_boolean(
    "PROFILE_CODE", False,
    description="Profile each test invocation")
#
# For use in tests
RUN_SLOW_TESTS = system.getenv_bool(
    "RUN_SLOW_TESTS", False,
    description="Run tests that can a while to run")
debug.reference_var(RUN_SLOW_TESTS)

UNDER_COVERAGE = system.getenv_bool(
    "COVERAGE_RUN", False,
    description="whether or not tests are being run under coverage"
)


# Dynamic imports
if PROFILE_CODE:
    import cProfile
    import pstats

#-------------------------------------------------------------------------------

def get_temp_dir(keep: Optional[bool] = None, unique=None) -> str:
    """Get temporary directory, omitting later deletion if KEEP
    Note: Optionally returns UNIQUE dir
    """
    # TODO2: rework keep using atexit-style callback
    debug.assertion(not ((keep is False) and (unique is False)))
    if keep is None:
        keep = KEEP_TEMP
    dir_base = gh.get_temp_file()
    if unique:
        # Note: creates parent temp dir if temp file regular file
        if not system.is_directory(dir_base):
            dir_base += "_temp_dir_"
            gh.full_mkdir(dir_base)
        dir_path = tempfile.NamedTemporaryFile(
            delete=(not keep), dir=dir_base).name
    else:
        dir_path = dir_base
    if not system.is_directory(dir_path):
        gh.full_mkdir(dir_path, force=True)
    debug.trace(5, f"unittest_wrapper.get_temp_dir() => {dir_path}")
    return dir_path


def trap_exception(function: Callable) -> Any:
    """Decorator to trap exception during function execution
    Note:
    - Only intended for use in tests (e.g., fix for maldito pytest).
    - Issues assertion so that test fails.
    - Should be inside any pytest.mark.xfail decorators.
    """
    debug.trace(8, f"trap_exception({gh.elide(function)}")
    #
    def wrapper(*args):
        """Wrapper around variable arity function f"""   ## TODO: {function.__name__}
        debug.trace(7, f"in wrapper: args={args}")
        result = None
        try:
            result = function(*args)
        except AssertionError:
            raise
        except:
            system.print_exception_info(function.__name__)
            assert(False)
        return result
    #
    debug.trace(7, f"trap_exception() => {gh.elide(wrapper)}")
    return wrapper


def pytest_fixture_wrapper(function: Callable) -> Callable:
    """Decorator for use with pytest fixtures like capsys
    Usage:
        @pytest_fixture_wrapper
        @trap_exception
        def test_it(capsys):
            ...
    """
    # See https://stackoverflow.com/questions/19614658/how-do-i-make-pytest-fixtures-work-with-decorated-functions
    # Note: This is currently usused. It was previously used with trap_exception.
    debug.trace(8, f"pytest_fixture_wrapper({gh.elide(function)}")
    #
    def wrapper(x):
        """Wrapper around unary function f(x)"""   ## TODO: {function.__name__}
        debug.trace(7, f"in wrapper: x={x}")
        return function(x)
    #
    debug.trace(7, f"pytest_fixture_wrapper() => {gh.elide(wrapper)}")
    return wrapper


def invoke_tests(filename: str, via_unittest: bool = VIA_UNITTEST):
    """Invoke TESTS defined in FILENAME, optionally VIA_UNITTEST"""
    debug.trace(5, f"invoke_tests({filename}, [{via_unittest}])")
    try:
        if via_unittest:
            unittest.main(argv=sys.argv[:1])
        else:
            pytest.main([filename])
    except:
        debug.raise_exception(6)
        system.print_exception_info("invoke_tests")


def init_temp_settings():
    """Initialize settings related to temp-file names"""
    ok = True
    # Re-initalize glue helper temp file settings
    ## TODO?: system.setenv("PRESERVE_TEMP_FILE", "1")
    debug.trace_expr(4, os.environ.get(PRESERVE_TEMP_FILE_LABEL))
    ## TEST: os.environ["PRESERVE_TEMP_FILE"] = "1"
    if not system.getenv(PRESERVE_TEMP_FILE_LABEL):
        system.setenv(PRESERVE_TEMP_FILE_LABEL, "1")
    gh.init()
    debug.trace_expr(4, os.environ.get(PRESERVE_TEMP_FILE_LABEL))
    return ok
        
#-------------------------------------------------------------------------------

class TestWrapper(unittest.TestCase):
    """Class for testcase definition
    Note:
    - script_module should be overriden to specify the module instance, such as via get_testing_module_name (see test/template.py)
    - set it to None to avoid command-line invocation checks
    """
    init_ok = init_temp_settings()

    script_file = TODO_FILE             # path for invocation via 'python -m coverage run ...' (n.b., usually set via get_module_file_path)
    script_module = TODO_MODULE         # name for invocation via 'python -m' (n.b., usually set via derive_tested_module_name)
    temp_base = TEMP_BASE
    check_coverage = system.getenv_bool(
        "CHECK_COVERAGE", False,
        desc="Check coverage during unit testing")
    ## TODO: temp_file = None
    ## TEMP: initialize to unique value independent of temp_base
    temp_file = None
    use_temp_base_dir = (system.is_directory(temp_base) if temp_base else False)
    ## OLD: test_num = 1
    test_num = 0
    temp_file_count = 0
    class_setup = False
    profiler = None
    monkeypatch = None
    capsys = None

    ## TEST:
    ## NOTE: leads to pytest warning. See
    ##   https://stackoverflow.com/questions/62460557/cannot-collect-test-class-testmain-because-it-has-a-init-constructor-from
    ## def __init__(self, *args, **kwargs):
    ##     debug.trace_fmtd(6, "TestWrapper.__init__({a}): keywords={kw}; self={s}",
    ##                      a=",".join(args), kw=kwargs, s=self)
    ##     super().__init__(*args, **kwargs)
    ##     debug.trace_object(7, self, label="TestWrapper instance")
    ##
    ## __test__ = False                 # make sure not assumed test
        
    @classmethod
    def setUpClass(cls, filename=None, module=None):
        """Per-class initialization: make sure script_module set properly
        Note: Optional FILENAME is path for testing script and MODULE the imported object for tested script
        """
        debug.trace(5, f"TestWrapper.setUpClass({cls}, fn={filename}, mod={module})")
        super().setUpClass()
        cls.class_setup = True
        debug.trace_object(7, cls, "init TestWrapper class")
        debug.assertion(cls.script_module != TODO_MODULE)
        if (cls.script_module is not None):
            # Try to pull up usage via python -m mezcla.xyz --help
            help_usage = gh.run("python -m '{mod}' --help", mod=cls.script_module)
            debug.assertion("No module named" not in help_usage,
                            f"problem running via 'python -m {cls.script_module}'")
            # Warn about lack of usage statement unless "not intended for command-line" type warning issued
            # TODO: standardize the not-intended wording
            if (not ((my_re.search(r"(warning|FYI):.*not intended", help_usage,
                                   flags=my_re.IGNORECASE))
                     or ("usage:" in help_usage.lower()))):
                system.print_stderr("Warning: script should implement --help")

        # Optionally, setup temp-base directory (normally just a file)
        debug.assertion(cls.init_ok)
        if cls.use_temp_base_dir is None:
            cls.use_temp_base_dir = gh.USE_TEMP_BASE_DIR
        if not cls.temp_base:
            cls.temp_base = cls.temp_file
        if not cls.temp_base:
            cls.temp_base = gh.get_temp_file()
        if cls.use_temp_base_dir:
            gh.full_mkdir(cls.temp_base, force=True)

        # Warn that coverage support is limited
        if cls.check_coverage:
            # note: For proper invocation info, see https://coverage.readthedocs.io/en/latest
            test_module = "pytest ..." if not VIA_UNITTEST else "unittest discover"
            debug.trace(4, "FYI: coverage check only covers run_script usages; "
                        "invoke externally for more general support:\n"
                        f"    python -m coverage run -m {test_module}")

        # Enable code profiling if desired
        if PROFILE_CODE:
            cls.profiler = cProfile.Profile()
            
        # Optionally, setup up script_file and script_module
        if filename:
            cls.set_module_info(filename, module_object=module)
        debug.trace_object(6, cls, "finalized TestWrapper class")

        return

    @staticmethod
    def derive_tested_module_name(test_filename: str) -> str:
        """Derive the name of the module being tested from TEST_FILENAME. Used as follows:
              script_module = TestWrapper.derive_tested_module_name(__file__)
        Note: *** Deprecated method *** (see get_testing_module_name)
        """
        debug.trace(3, "Warning: in deprecrated derive_tested_module_name")
        debug.trace(5, f"TestWrapper.derive_tested_module_name({test_filename})")
        module = os.path.split(test_filename)[-1]
        module = my_re.sub(r".py[oc]?$", "", module)
        module = my_re.sub(r"^test_", "", module)
        debug.trace_fmtd(5, "derive_tested_module_name({f}) => {m}",
                         f=test_filename, m=module)
        return (module)

    @staticmethod
    def get_testing_module_name(test_filename: str, module_object: Optional[object] = None) -> str:
        """Derive the name of the module being tested from TEST_FILENAME and MODULE_OBJECT
        Note: used as follows (see tests/template.py):
            script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
        """
        debug.trace(6, f"TestWrapper.get_testing_module_name({test_filename}, {module_object})")
        # Note: Used to resolve module name given THE_MODULE (see template).
        module_name = os.path.split(test_filename)[-1]
        module_name = my_re.sub(r".py[oc]?$", "", module_name)
        module_name = my_re.sub(r"^test_", "", module_name)
        package_name = THIS_PACKAGE
        if module_object is not None:
           package_name = getattr(module_object, "__package__", "")
           debug.trace_expr(4, package_name)
        if package_name:
            full_module_name = package_name + "." + module_name
        else:
            full_module_name = module_name
        debug.trace_fmtd(4, "get_testing_module_name({f}, [{mo}]) => {m}",
                         f=test_filename, m=full_module_name, mo=module_object)
        return (full_module_name)

    @staticmethod
    def get_module_file_path(test_filename: str) -> str:
        """Return absolute path of module being tested"""
        result = system.absolute_path(test_filename)
        ## TODO3: use os.path.delim instead of /
        result = my_re.sub(r'tests\/test_(.*\.py)', r'\1', result)
        debug.assertion(result.endswith(".py"))
        debug.trace(6, f'get_module_file_path({test_filename}) => {result}')
        return result

    @classmethod
    def set_module_info(cls, test_filename: str, module_object: Optional[object] = None) -> None:
        """Sets both script_module and script_path
        Note: normally invoked in setUpClass method
        Usage: cls.set_module_info(__file__, THE_MODULE)
        """
        debug.trace(7, f'set_module_info({test_filename}, {module_object})')
        cls.script_module = cls.get_testing_module_name(test_filename, module_object)
        cls.script_file = cls.get_module_file_path(test_filename)
        return 
    
    def ensure_file_dir_exists(self, filename):
        """Make sure that directory for FILENAME exists"""
        debug.trace(5, f"TestWrapper.ensure_file_dir_exists({filename})")
        dir_path = gh.dirname(filename)
        if not system.is_directory(dir_path):
            debug.trace(4, f"FYI: Creating output directory {dir_path!r}\n\t{self.temp_file=}")
            gh.full_mkdir(dir_path)

    def setUp(self) -> None:
        """Per-test initializations
        Notes:
        - Disables tracing scripts invoked via run() unless ALLOW_SUBCOMMAND_TRACING
        - Initializes temp file name (With override from environment)."""
        # Note: By default, each test gets its own temp file.
        debug.trace(5, "TestWrapper.setUp()")
        if not self.class_setup:
            debug.trace(3, "Warning: invoking setUpClass in setUp; make sure seUpClass calls parent")
            TestWrapper.setUpClass(self.__class__)
        if not gh.ALLOW_SUBCOMMAND_TRACING:
            gh.disable_subcommand_tracing()
        # The temp file is an extension of temp-base file by default.
        # Optionally, if can be a file in temp-base subdrectory.
        if self.use_temp_base_dir:
            default_temp_file = gh.form_path(self.temp_base, "test-")
        else:
            default_temp_file = self.temp_base + "-test-"
        temp_file_basename = default_temp_file
        TestWrapper.test_num += 1
        default_temp_file += str(TestWrapper.test_num)
        debug.trace_expr(5, default_temp_file)

        # Get new temp file and delete existing file and variants based on temp_file_count,
        # such as /tmp/test-2, /tmp/test-2-1, and /tmp/test-2-2 (but not /tmp/test-[13]*).
        # Warning: using TEMP_FILE is not recommended due to clobbering by different tests
        self.temp_file = (gh.TEMP_FILE or default_temp_file)
        if PRUNE_TEMP:
            gh.delete_existing_file(f"{self.temp_file}")
            for f in gh.get_matching_files(f"{temp_file_basename}-[0-9]*"):
                gh.delete_existing_file(f)

        # Start the profiler
        if PROFILE_CODE:
            assert self.profiler is not None
            self.profiler.enable()
        
        debug.trace_object(6, self, "TestWrapper instance")
        return

    def run_script(
            self,
            options: Optional[str] = None,
            data_file: Optional[str] = None,
            log_file: Optional[str] = None,
            trace_level: IntOrTraceLevel = 4,
            out_file: Optional[str] = None,
            env_options: Optional[str] = None,
            uses_stdin: Optional[bool] = None,
            post_options: Optional[str] = None,
            background: Optional[str] = None,
            skip_stdin: Optional[bool] = None
        ) -> str:
        """Runs the script over the DATA_FILE (optional), passing (positional)
        OPTIONS and optional setting ENV_OPTIONS. If OUT_FILE and LOG_FILE are
        not specified, they  are derived from self.temp_file. The optional POST_OPTIONS
        go after the data file.
        Notes:
        - OPTIONS uses quotes around shell special characters used (e.g., '<', '>', '|')
        - issues warning if script invocation leads to error
        - if USES_STDIN, requires explicit empty string for DATA_FILE to avoid use of - (n.b., as a precaution against hangups)
        - if SKIP_STDIN, then - omitted from command line
        - By default, stderr is not included in the output
        """
        debug.trace_fmtd(trace_level + 1,
                         "TestWrapper.run_script(opts={opts!r}, data={df}, log={lf}, lvl={lvl}, out={of}, env={env}, stdin={stdin}, post={post}, back={back})",
                         opts=options, df=data_file, lf=log_file, lvl=trace_level, of=out_file,
                         env=env_options, stdin=uses_stdin, post=post_options, back=background)
        if options is None:
            options = ""
        if env_options is not None:
            suffix = ' '
            preffix = ' '
            if os.name == 'nt':
                suffix = '&& '
                preffix = 'SET '
            list_options = [ f"{preffix}{option}{suffix}" for option in env_options.split(' ')]
            env_options = " ".join(list_options)
        else:
            env_options = ""
        if post_options is None:
            post_options = ""
        if skip_stdin is None:
            skip_stdin = False

        # Derive the full paths for data file and log, and then invoke script.
        # TODO: derive from temp base and data file name?;
        # TODO1: derive default for uses_stdin based on use of filename argment (e.g., from usage)
        uses_stdin_false = ((uses_stdin is not None) and not bool(uses_stdin))
        data_path = ("" if (skip_stdin or uses_stdin_false) else "-")
        if data_file is not None:
            data_path = (gh.resolve_path(data_file) if len(data_file) else data_file)
        assert isinstance(self.temp_file, str)
        if not log_file:
            log_file = self.temp_file + ".log"
        if not out_file:
            out_file = self.temp_file + ".out"
        # note: output is redirected to a file to preserve tabs

        # Make sure output and log files are in valid dirs
        self.ensure_file_dir_exists(out_file)
        self.ensure_file_dir_exists(log_file)

        # Set converage script path and command spec
        coverage_spec = ''
        script_module = self.script_module
        if self.check_coverage:
            debug.assertion(self.script_file)
            script_module = self.script_file
            coverage_spec = 'coverage run'
        debug.assertion(not script_module.endswith(".py"))
        amp_spec = "&" if background else ""

        # Run the command
        ## TODO3: allow for stdin_command (e.g., "echo hey" | ...)
        ## TODO2: add sanity check for special shell characters
        ##   shell_tokens = ['<', '>', '|']
        ##   debug.assertion(not system.intersection(options.split(), shell_tokens))
        gh.issue("{env} python -m {cov_spec} {module}  {opts}  {path}  {post} 1> {out} 2> {log} {amp_spec}",
                 env=env_options, cov_spec=coverage_spec, module=script_module,
                 opts=options, path=data_path, out=out_file, log=log_file, post=post_options, amp_spec=amp_spec)
        output = system.read_file(out_file)
        # note: trailing newline removed as with shell output
        if output.endswith("\n"):
            output = output[:-1]
        debug.trace_fmtd(trace_level, "output: {{\n{out}\n}}",
                         out=gh.indent_lines(output), max_len=4096)

        # Make sure no python or bash errors. For example,
        #   "SyntaxError: invalid syntax" and "bash: python: command not found"
        log_contents = system.read_file(log_file)
        error_found = my_re.search(r"(\S+error:)|(no module)|(command not found)",
                                   log_contents.lower())
        debug.assertion(not error_found)
        debug.trace_expr(trace_level + 1, log_contents, max_len=4096)

        # Do sanity check for python exceptions
        traceback_found = my_re.search("Traceback.*most recent call", log_contents)
        debug.assertion(not traceback_found)

        return output

    def resolve_assertion(
            self,
            function_label: str,
            message: Optional[str]
        ) -> Tuple[
            Optional[str],
            Optional[str],
            Optional[int],
            Optional[str],
            Optional[str]
        ]:
        """Returns statement text, filename, line number, and qualifier for FUNCTION_LABEL assertion failure"""
        ## TODO2: use new introspection module
        statement = filename = line_num = expr = qual = None
        try:
            # note: accounts for trap_exception and other decorators
            for caller in inspect.stack():
                debug.trace_expr(8, caller)
                ## DEBUG: print(f"{caller=}")
                (_frame, filename, line_num, _function, context, _index) = caller
                statement = debug.read_line(filename, line_num).strip()
                ## DEBUG: print(f"{statement=}")
                if f".{function_label}(" in statement:
                    break
            # TODO3: use abstract syntax tree (AST) based extraction
            # ex: self.do_assert(not my_re.search(r"cat|dog", description))  # no pets
            # Isolate condition
            cond = my_re.sub(fr"^\s*\S+\.{function_label}\((.*)\)", r"\1", statement)
            # Get expression proper, removing optional comments and semicolon 
            expr = my_re.sub(r";?\s*#.*$", "", cond)
            # Strip optional message
            qual = ""
            if message is not None:
                expr = my_re.sub(r", *([\'\"]).*\1\s*$", "", expr)   # string arg
                expr = my_re.sub(r", *[a-z0-9_]+$", "", expr,        # variable arg
                                 flags=my_re.IGNORECASE)
                qual = f": {message}"
            debug.trace_expr(7, filename, line_num, context, prefix="resolve_assertion: ")
        except:
            system.print_exception_info("resolve_assertion")
        debug.assertion(statement)
        result = (statement, filename, line_num, expr, qual)
        debug.trace(7, f"resolve_assertion({function_label}) => {result}")
        return result
    
    def do_assert(
            self,
            condition: bool,
            message: Optional[str] = None
        ) -> None:
        """Shows context for assertion failure with CONDITION and then issue assert
        If MESSAGE specified, included in assertion error
        Note:
        - Works around for maldito pytest, which makes it hard to do simple things like pinpointing errors.
        - Formatted similar to debug.assertion:
             Test assertion failed: <expr> (at <f><n>): <msg>
        """
        debug.trace(7, f"do_assert({condition}, msg={message})")
        if ((not condition) and debug.debugging(debug.TL.DEFAULT)):
            (statement, filename, line_num, expr, qual) = self.resolve_assertion("do_assert", message)
            if statement:
                # Format assertion error with optional qualification (i.e., user message)
                debug.trace(1, f"Test assertion failed: {expr} (at {filename}:{line_num}){qual}")
                debug.trace(5, f"\t{statement}")
            else:
                system.print_error("Warning: unexpected condition in do_assert")
        assert condition, message
    #
    ## TODO:
    ## assert = do_assert
    ##
    ## TEST:
    ## def assert(self, *args, **kwargs):
    ##     """Wrapper around do_assert (q.v.)"""
    ##     self.do_assert(*args, **kwargs)

    def do_assert_equals(
            self,
            value1: Any,
            value2: Any,
            message: Optional[str] = None
        ) -> None:
        """Make sure VALUE1 equals VALUE2, using optional MESSAGE"""
        equals = value1 == value2
        if ((not equals) and debug.debugging(debug.TL.DEFAULT)):
            (statement, filename, line_num, expr, qual) = self.resolve_assertion("do_assert_equals", message)
            if statement:
                # Format assertion error with optional qualification (i.e., user message)
                debug.trace(1, f"Test equality assertion failed: {expr} (at {filename}:{line_num}){qual}")
                debug.trace(5, f"\t{statement}")
                debug.trace(2, "diff:\n" + string_diff(value1, value2))
            else:
                system.print_error("Warning: unexpected condition in do_assert_equals")
        assert equals, message
    
    @pytest.fixture(autouse=True, name='monkeypatch')
    def monkeypatch_fixture(self, monkeypatch) -> None:
        """Support for using pytest monkeypatch to modify objects (e.g., dictionaries or environment variables)"""
        # See https://docs.pytest.org/en/latest/how-to/monkeypatch.html
        self.monkeypatch = monkeypatch

    @pytest.fixture(autouse=True, name='capsys')
    def capsys_fixture(self, capsys) -> None:
        """Support for capturing stdout and stderr"""
        # See https://docs.pytest.org/en/latest/how-to/capture-stdout-stderr.html
        self.capsys = capsys

    def get_stdout_stderr(self) -> Tuple[str, str]:
        """Get currently captured standard output and error
        Note: Clears both stdout and stderr captured afterwards. This might
        be needed beforehand to clear capsys buffer.

        Warning: The capsys clearing workaround is not foolproof, so you
        might need to disable capsys beforehand (e.g., in setUp): see
        tests/template.py. Alternatively, use multiple test functions:
            https://stackoverflow.com/questions/56187165/how-to-clear-captured-stdout-stderr-in-between-of-multiple-assert-statements
        """
        stdout, stderr = ("", "")
        try:
            with self.capsys.disabled():
                stdout, stderr = self.capsys.readouterr()
                ## TODO4: resolve issue with resolve_assertion call-stack tracing being clippped
                debug.trace_expr(5, stdout, stderr, prefix="get_stdout_stderr:\n", delim="\n", max_len=16384)
        except:
            # note: trace level high so as not to affect normal testing
            debug.trace_exception(7, "get_stdout_stderr")
        return stdout, stderr
        
    def get_stdout(self) -> str:
        """Get currently captured standard output (see get_stdout_stderr)
        Warning: You might need to invoke beforehand to clear buffer.
        """
        stdout, _stderr = self.get_stdout_stderr()
        return stdout
        
    def get_stderr(self) -> str:
        """Get currently captured standard error (see get_stdout_stderr)
        Warning: You might need to invoke beforehand to clear buffer.
        """
        _stdout, stderr = self.get_stdout_stderr()
        return stderr

    def clear_stdout_stderr(self) -> None:
        """Clears stdout and stderr by issuing dummy call.
        Warning: See disclaimer under get_stdout_stderr.
        """
        _result = self.get_stdout_stderr()
        return
    #
    clear_stdout = clear_stdout_stderr
    clear_stderr = clear_stdout_stderr
    
    def get_temp_file(self, delete: Optional[bool] = None, static: Optional[bool] = None) -> str:
        """Return name of temporary file based on self.temp_file, optionally with DELETE.
        Normally, the file is based on the test base, current test number and usage count
        (e.g., /tmp/_temp-fi7huvmb_-test-1-3 for third temp file used in first test).
        Note:
        - This returns a new file each time called unless STATIC specified.
        - The delete option is not yet functional.
        """
        # Note: delete defaults to False if detailed debugging
        # TODO: allow for overriding other options to NamedTemporaryFile
        if delete is None and debug.detailed_debugging():
            delete = False
        temp_file_name = self.temp_file
        if not static:
            temp_file_name += f"-{self.temp_file_count}"
        self.temp_file_count += 1
        debug.assertion(not delete, "Support for delete not implemented")
        debug.trace(5, f"TestWrapper.get_temp_file() => {temp_file_name!r}")
        return temp_file_name

    def get_temp_dir(self, delete: Optional[bool] = None,
                     skip_create: Optional[bool] = None,
                     static: Optional[bool] = None) -> str:
        """Return name of temporary dir based on self.temp_file, optionally with DELETE.
        Also, the directory will be created unless SKIP_CREATE;
        this possibly overwrites existing file with same name.
        In addition, the directory will be unique unless STATIC specified.
        Note: delete option not yet functional
        """
        temp_dir = self.get_temp_file(delete=delete, static=static)
        if not skip_create:
            gh.full_mkdir(temp_dir, force=True)
        debug.trace(5, f"get_temp_dir() => {temp_dir!r}")
        return temp_dir
    
    def create_temp_file(self, contents: Any,  binary: bool = False) -> str:
        """Create temporary file with CONTENTS and return full path"""
        temp_filename = self.get_temp_file()
        system.write_file(temp_filename, contents, binary=binary)
        debug.trace(6, f"create_temp_file({contents!r}) => {temp_filename}")
        return temp_filename

    def patch_trace_level(self, level):
        """Monkey patch the trace LEVEL"""
        self.monkeypatch.setattr("mezcla.debug.trace_level", level)

    def tearDown(self) -> None:
        """Per-test cleanup: deletes temp file unless detailed debugging"""
        debug.trace(6, "TestWrapper.tearDown()")
        if not KEEP_TEMP:
            gh.run("rm -vf {file}*", file=self.temp_file)
            for i in range(self.temp_file_count):
                gh.run(f"rm -vf {self.temp_file}-{i}")
        self.temp_file_count = 0

        # Show results of code profiling if enabled
        if PROFILE_CODE:
            assert self.profiler is not None
            self.profiler.disable()
            ## TODO: debug.trace(1, f"Test {self.test_num} code profiling results")
            print(f"Test {self.test_num} code profiling results")
            ## OLD: self.profiler.print_stats(sort='cumulative')
            ## NOTE: Based on cProfile's print_stats (in order to use stderr)
            stats = pstats.Stats(self.profiler, stream=sys.stderr)
            stats.strip_dirs().sort_stats('cumulative').print_stats()
        return

    @classmethod
    def tearDownClass(cls) -> None:
        """Per-class cleanup: stub for tracing purposes"""
        debug.trace_fmtd(5, "TestWrapper.tearDownClass(); cls={c}", c=cls)
        if not KEEP_TEMP:
            ## TODO: use shutil
            if cls.use_temp_base_dir:
                if DISABLE_RECURSIVE_DELETE:
                    debug.trace(4, f"FYI: Only deleting top-level files in {cls.temp_base} to avoid potentially dangerous rm -r")
                    gh.run("rm -f {dir}/* {dir}/.*", dir=cls.temp_base)
                    gh.run("rm -f {dir}", dir=cls.temp_base)
                else:
                    debug.trace(4, f"FYI: Using potentially dangerous rm -r over {cls.temp_base}")
                    gh.run("rm -rvf {dir}", dir=cls.temp_base)
            else:
                gh.run("rm -vf {base}*", base=cls.temp_base)
        super().tearDownClass()
        return

## TODO: TestWrapper.assert = TestWrapper.do_assert

#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    debug.trace_current_context(level=TL.QUITE_DETAILED)
    debug.trace(TL.USUAL, "Warning: not intended for command-line use\n")



#! /usr/bin/env python3
#
# Tests for train_language_model module
#
# Notes:
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_train_language_model.py
#
# TODO2:
# - Check for other cases of test classes without tests (due to work-in-progress nature
#   prior to 2025).
# - Likewise check for tests not using tests/template.py, where there are to-do tests with xfail,
#   allowing for easier tracking of tests to be implemeted!
#
#...............................................................................
# Sample output tested (2grams over LICENSE.txt):
#
#   \2-grams:
#   \data\
#   ngram 1=319
#   ngram 2=812
#   
#   \1-grams:
#   ...
#   -2.597798	gnu	-0.76723224
#   ...
#
#   \2-grams:
#   ...
#   -0.05367994	general public
#   ...
#   -0.15376821	combined work
#   ...
#
# Note:
#   $ count_it.perl -i '\w+ \w+' ~/Mezcla/LICENSE.txt | head
#   combined work	13
#   ...
#   general public	6
#   public license	5
#   ...
#
#................................................................................
# Sample usage:
#
# $ train_language_model.py --usage-notes
# /home/tomohara/Mezcla/mezcla/train_language_model.py --usage-notes
# 
# Usage: /home/tomohara/Mezcla/mezcla/train_language_model.py source-file.txt
#
# ...
#
# Notes:
# - The output is put in source-file.Ngram.arpa and source-file.Ngram.mmap.
# 

"""Tests for train_language_model module"""

# Standard modules
## NOTE: this is empty for now

# Installed modules
import pytest

# Local modules
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.my_regex import my_re
from mezcla import system
from mezcla.unittest_wrapper import TestWrapper, invoke_tests

# Note: Rreference are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.train_language_model as THE_MODULE

class TestTrainLanguageModel(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    license_path = gh.resolve_path("LICENSE.txt", heuristic=True)

    def get_score(self, ngram):
        """Return score for NGRAM from .arpa file"""
        # Note: See format above
        result = system.to_float(gh.run(f"grep '{ngram}' {self.temp_file}.arpa | cut -f1"))
        debug.trace(5, f"get_score({ngram}) => {result}")
        return result
        
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_01_data_file(self):
        """Tests run_script w/ data file"""
        # Warning: see notes above about potential issues with run_script-based tests.
        debug.trace(4, f"TestIt.test_01_data_file(); self={self}")
        self.run_script(
            env_options="MAX_NGRAM=2",
            options=f"--tokenize --output-basename {self.temp_file}",
            data_file=self.license_path)
        #
        combined_work_score = self.get_score("combined work")
        general_public_score = self.get_score("general public")
        self.do_assert(combined_work_score > general_public_score)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_02_something_else(self):
        """Test usage statement"""
        debug.trace(4, f"TestIt.test_02_usage(); self={self}")
        THE_MODULE.usage()
        output = self.get_stdout()
        self.do_assert(my_re.search("output.*source-file.Ngram.arpa", output))
        return


if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Test(s) for ../randomize_lines.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_randomize_lines.py
#

"""Tests for randomize_lines module"""

# Standard packages
## NOTE: this is empty for now

# Installed packages
import pytest

# Local packages
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla.unittest_wrapper import trap_exception
from mezcla import debug
from mezcla import system
from mezcla import tpo_common as tpo

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.randomize_lines as THE_MODULE

class TestRandomizeLines(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    @pytest.mark.xfail                   # TODO: remove xfail
    @trap_exception
    def test_data_file(self):
        """Tests run_script w/ data file"""
        debug.trace(4, f"TestIt.test_data_file(); self={self}")
        NUM_LINES = 100
        data = [f"line {l}" for l in range(NUM_LINES)]
        system.write_lines(self.temp_file, data)
        ## TODO: add use_stdin=True to following if no file argument
        output = self.run_script(options="--percent 10", data_file=self.temp_file)
        random_lines = output.splitlines()
        self.do_assert(len(random_lines) == 10)
        self.do_assert(tpo.is_subset(random_lines, data))
        return


if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Test(s) for gensim_test.py
#
# Notes:
# - Fill out the TODO's through the file.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_gensim_test.py
#
#------------------------------------------------------------------------
# TODO: adds tests for following
# $ DEBUG_LEVEL=4  /usr/bin/time python  -u  -m gensim_test  --save  --tfidf  --verbose  random100-titles-descriptions.txt  >| random100-titles-descriptions.log  2>&1
# $ grep food random100-titles-descriptions.verbose.log | less -S
# (3, [('food', 0.5934840534383484), ('accounting', 0.21278874808559092), ('beverage', 0.20959194756482724),
# (53, [('general', 0.3528312412285222), ('manager', 0.3148377686034749), ('food', 0.27899845787097777),
#

"""Tests for gensim_test module"""

# Standard packages
import re

# Installed packages
import pytest
## TAKE1
## # note: The gensim module is not installed by default, so tests skipped if not found
## try:
##     import gensim
## except:
##     gensim = None

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla import tpo_common as tpo
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla import system

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
try:
    ## OLD: # pylint: disable=ungrouped-imports
    import gensim
    import mezcla.gensim_test as THE_MODULE
except:
    ## TEST (maldito pytest):
    ## print("hey, hey, hey!")
    debug.trace_exception(3, "importing gensim_test")
    gensim = None
    THE_MODULE = None

class TestGensimTest(TestWrapper):
    """Class for script-level testcase definition"""
    script_file = TestWrapper.get_module_file_path(__file__)
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    @pytest.mark.skipif(not gensim, reason="gensim module missing")
    def test_data_file(self):
        """Tests results over a known data file (LICENSE.txt)"""
        tpo.debug_print("test_data_file()", 4)
        data_file = "LICENSE.txt"
        temp_data_file = self.temp_base + "-" + data_file
        gh.copy_file(gh.resolve_path(data_file), temp_data_file)
        output = self.run_script("--save", temp_data_file)
        ## TODO: assert re.search("storing corpus in Matrix Market format", output)
        assert gh.non_empty_file(temp_data_file.replace(".txt", ".bow.mm"))
        debug.trace_expr(5, output)
        ## TODO3: check for expected entries (e.g., .wordids.txt.bz2 file)
        return

    @pytest.mark.skipif(not gensim, reason="gensim module missing")
    def test_vector_printing(self):
        """Test printing of corpus vector for simple input"""
        tpo.debug_print("test_vector_printing()", 4)
        temp_file = self.temp_base + ".txt"
        system.write_file(temp_file, "My dog has fleas.\n")
        output = self.run_script("--print --verbose", temp_file)
        assert re.search(r"\(u?'dog', 1\),.*\(u?'has', 1\)", output)
        ## TODO4: likewise do non-trivial input like LICENSE.txt
        debug.trace_expr(5, output)
        return

    
class TestGensimTest2:
     """Class for internal testcase definitions"""

     @pytest.mark.skipif(not gensim, reason="gensim module missing")
     def test_corpus_data(self):
         """Test creation of corpus from a file"""
         corpus = THE_MODULE.CorpusData(__file__)
         # note: currently 81 unique tokens extracted
         assert len(list(corpus)) > 50
   
#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
#
# via https://github.com/Paperspace/training_styletransfer

"""A script to dump tensorflow checkpoint variables to deeplearnjs.
This script takes a checkpoint file and writes all of the variables in the
checkpoint to a directory.
"""

# TEMP: pylint: disable=unspecified-encoding

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

## OLD: import argparse
import json
import os
import re
import string
## OLD: import tensorflow as tf
## TODO?:
import tensorflow.compat.v1 as tf

# Local modules
# TODO: def mezcla_import(name): ... components = eval(name).split(); ... import nameN-1.nameN as nameN
from mezcla import debug
from mezcla.main import Main
from mezcla import system

# Command-line arguments
JSON_OUTPUT = "json"
CHECKPOINT_FILE = "checkpoint"
OUTPUT_DIR = "output-dir"


## TODO:
## # Environment options
## # Note: These are just intended for internal options, not for end users.
## # It also allows for enabling options in one place rather than four
## # (e.g., [Main member] initialization, run-time value, and argument spec., along
## # with string constant definition).
## #
## FUBAR = system.getenv_bool("FUBAR", False,
##                            description="Fouled Up Beyond All Recognition processing")


FILENAME_CHARS = string.ascii_letters + string.digits + '_'

#-------------------------------------------------------------------------------

def _var_name_to_filename(var_name):
    """Return filename to use fo VAR_NAME"""
    chars = []
    for c in var_name:
        if c in FILENAME_CHARS:
            chars.append(c)
        elif c == '/':
            chars.append('_')
        else:
            debug.trace(7, f"Excluding char '{c}' from filename")
    return ''.join(chars)

def remove_optimizer_variables(output):
    """Remvoes tensorflow variables used for optimization"""
    ## OLD: vars_dir = os.path.expanduser(output)
    manifest_file = os.path.join(output, 'manifest.json')
    with open(manifest_file) as f:
        manifest = json.load(f)
        new_manifest = {key: manifest[key] for key in manifest 
                        if 'Adam' not in key and 'beta' not in key}
    with open(manifest_file, 'w') as f:
        json.dump(new_manifest, f, indent=2, sort_keys=True)

    for name in os.listdir(output):
        if 'Adam' in name or 'beta' in name:
            os.remove(os.path.join(output, name))


def dump_checkpoints(checkpoint_dir, output, use_json=False):
    """Dump all tensorflow variabes in CHECKPOINT_DIR [file???] to OUTPUT dir.
    Note: Optionally USEs_JSON format for arrays"""
    chk_fpath = os.path.expanduser(checkpoint_dir)
    reader = tf.train.NewCheckpointReader(chk_fpath)
    var_to_shape_map = reader.get_variable_to_shape_map()
    output_dir = os.path.expanduser(output)
    tf.gfile.MakeDirs(output_dir)
    manifest = {}
    ## TODO: remove_vars_compiled_re = re.compile('')

    var_filenames_strs = []
    for name in var_to_shape_map:
        ## OLD:
        ## if ('' and
        ##     re.match(remove_vars_compiled_re, name)) or name == 'global_step':
        if name == 'global_step':
            debug.trace(5, f"Ignoring tf var {name}")
            continue
        var_filename = _var_name_to_filename(name)
        manifest[name] = {'filename': var_filename, 'shape': var_to_shape_map[name]}

        tensor = reader.get_tensor(name)
        if use_json:
            with open(os.path.join(output_dir, var_filename + ".json"), 'w') as f:
                f.write(json.dumps(tensor.tolist()))
        else:
            with open(os.path.join(output_dir, var_filename), 'wb') as f:
                f.write(tensor.tobytes())

    var_filenames_strs.append("\"" + var_filename + "\"")

    manifest_fpath = os.path.join(output_dir, 'manifest.json')
    print('Writing manifest to ' + manifest_fpath)
    with open(manifest_fpath, 'w') as f:
        f.write(json.dumps(manifest, indent=2, sort_keys=True))

    remove_optimizer_variables(output_dir)

#...............................................................................

class Script(Main):
    """Adhoc script class (e.g., no I/O loop, just run calls)"""
    json_output = False
    checkpoint_file = None
    output_dir = None

    def setup(self):
        """Check results of command line processing"""
        debug.trace_fmtd(5, "Script.setup(): self={s}", s=self)
        self.json_output = self.get_parsed_option(JSON_OUTPUT, self.json_output)
        self.checkpoint_file = self.get_parsed_argument(CHECKPOINT_FILE)
        self.output_dir = self.get_parsed_argument(OUTPUT_DIR)
        debug.trace_object(5, self, label="Script instance")

    def run_main_step(self):
        """Main processing step"""
        debug.trace_fmtd(5, "Script.run_main_step(): self={s}", s=self)
        dump_checkpoints(self.checkpoint_file, self.output_dir,
                         use_json=self.json_output)

#-------------------------------------------------------------------------------


if __name__ == '__main__':
    debug.trace_current_context(level=debug.QUITE_DETAILED)
    debug.trace_fmt(4, "Environment options: {eo}",
                    eo=system.formatted_environment_option_descriptions())
    app = Script(
        description=__doc__,
        skip_input=True,
        manual_input=True,
        boolean_options=[(JSON_OUTPUT, "Use JSON format for arrays, not binary")],
        # Note: FILENAME is default argument unless skip_input
        positional_arguments=[CHECKPOINT_FILE, OUTPUT_DIR], 
        # Note: Following added for indentation: float options are not common
        float_options=None)
    app.run()
    debug.assertion(not any(re.search(r"^TODO_", m, re.IGNORECASE)
                            for m in dir(app)))
                   



#! /usr/bin/env python3
#
# TODO: Test(s) for ../perform_regression.py
#
# Notes:
# - It would be a good idea to compare results versus commerial statistics
#   package (or at least Excel).
# - This can be run as follows (e.g., from root of repo):
#   $ pytest ./mezcla/adhoc/tests/test_perform_regression.py
#................................................................................
# Sample tested output (see input data below):
#
#   Call:
#   lm(formula = height ~ weight, data = data)
#   
#   Residuals:
#         Min        1Q    Median        3Q       Max 
#   -0.021355 -0.007024  0.002112  0.011106  0.014011 
#   
#   Coefficients:
#                Estimate Std. Error t value Pr(>|t|)    
#   (Intercept) 0.6484604  0.0292218   22.19 1.02e-11 ***
#   weight      0.0161443  0.0004679   34.50 3.60e-14 ***
#   ---
#   Signif. codes:  0 ‘***’ 0.001 ‘**’ 0.01 ‘*’ 0.05 ‘.’ 0.1 ‘ ’ 1
#   
#   Residual standard error: 0.01232 on 13 degrees of freedom
#   Multiple R-squared:  0.9892,	Adjusted R-squared:  0.9884 
#   F-statistic:  1190 on 1 and 13 DF,  p-value: 3.604e-14
#

"""Tests for perform_regression module"""

# Standard modules
## TODO: from collections import defaultdict

# Installed modules
import pytest

# Local modules
from mezcla.unittest_wrapper import TestWrapper
## TODO: from mezcla.unittest_wrapper import trap_exception
from mezcla import debug
## TODO: from mezcla import glue_helpers as gh
from mezcla.my_regex import my_re
from mezcla import system

# Note: Two references are used for the module to be tested:
#    THE_MODULE:                        global module object
#    TestIt.script_module:              path to file
THE_MODULE = None
try:
    import mezcla.adhoc.perform_regression as THE_MODULE
except:
    system.print_exception_info("perform_regression import") 

## TODO:
## # Environment options
## # Note: These are just intended for internal options, not for end users.
## # It also allows for enabling options in one place.
## #
## FUBAR = system.getenv_bool(
##     "FUBAR", False,
##     description="Fouled Up Beyond All Recognition processing")

# Sample data
# Height in meters vs. weight in kilograms
# via https://en.wikipedia.org/wiki/Simple_linear_regression
HEIGHT = "height 1.47 1.50 1.52 1.55 1.57 1.60 1.63 1.65 1.68 1.70 1.73 1.75 1.78 1.80 1.83".split()
WEIGHT = "weight 52.21 53.12 54.48 55.84 57.20 58.57 59.93 61.29 63.11 64.47 66.28 68.10 69.92 72.19 74.46".split()
debug.assertion(len(HEIGHT) == len(WEIGHT))

#------------------------------------------------------------------------

class TestIt(TestWrapper):
    """Class for command-line based testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    #
    # TODO: use_temp_base_dir = True            # treat TEMP_BASE as directory
    # note: temp_file defined by parent (along with script_module, temp_base, and test_num)

    @pytest.mark.xfail                   # TODO: remove xfail
    ## DEBUG: @trap_exception            # TODO: remove when debugged
    def test_01_data_file(self):
        """Tests run_script w/ data file"""
        # Warning: see notes above about potential issues with run_script-based tests.
        debug.trace(4, f"TestIt.test_01_data_file(); self={self}")
        data = [f"{h},{WEIGHT[i]}" for i, h in enumerate(HEIGHT)]
        system.write_lines(self.temp_file, data)
        ## TODO: add use_stdin=True to following if no file argument
        output = self.run_script(options="", data_file=self.temp_file)
        self.do_assert(my_re.search(r"R-squared.*0.9", output.strip()))
        self.do_assert(my_re.search(r"Intercept.*\*\*\*", output.strip()))
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    ## DEBUG: @trap_exception            # TODO: remove when debugged
    def test_02_something_else(self):
        """Test for something_else: TODO..."""
        debug.trace(4, f"TestIt.test_02_something_else(); self={self}")
        self.do_assert(False, "TODO: implement")
        ## ex: self.do_assert(THE_MODULE.TODO_function() == TODO_value)
        return

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    ## TODO2: here and elsewhere: invoke_tests(__file__)
    pytest.main([__file__])



#! /usr/bin/env python3
#
# Utility functions for writing glue scripts, such as implementing functionality
# available in Unix scripting (e.g., basename command).
#
# NOTE:
# - Some of the utilities are specific to Unix (e.g., full_mkdir and real_path). In
#   contrast, system.py attempts to be more cross platform.
# - ** It can be confusing debugging script that use run, because the trace level
#  is raised by default. To disable this, set the SUB_DEBUG_LEVEL as follows:
#     l=5; DEBUG_LEVEL=$l SUB_DEBUG_LEVEL=$l merge_files.py ...
# - Also see ALLOW_SUBCOMMAND_TRACING usage below and in unittest_wrapper.py.
# - By default, temporary files are created in the system temporary directory. To
#   facilate debugging, two environment variables allow for overriding this
#      TEMP_FILE: fixed temporary file to use (** avoid if possible)
#      TEMP_BASE: basename for temporary files (or a directory if USE_TEMP_BASE_DIR)
# - Uses str as return type instead of Optional[str] for string functions like basename
#   or get_temp_dir where "" is an invalid value (e.g., no need for None).
# - Avoids adding assert for mypy unless trapped.
#
# TODO:
# - Add more functions to facilitate command-line scripting (check bash scripts for commonly used features).
# - Add functions to facilitate functional programming (e.g., to simply debugging traces).
# TODO3: add deprecated warnings to functions superceded by ones in system
#

"""Helpers gluing scripts together

Usage example:

   cat {script} | python -c 'from mezcla import glue_helpers, system; print("\\n".join(glue_helpers.elide_values(system.read_lines("{script}"))))'
"""

# Standard packages
from collections import defaultdict
import glob
import inspect
import os
from pathlib import Path
import re
import shutil
from subprocess import getoutput
import sys
import tempfile
from typing import (
    Optional, Any, List, Dict, Union,
    ## OLD: TextIO,
)
from types import FrameType
## OLD: from io import TextIOWrapper
## DEBUG: sys.stderr.write(f"{__file__=}\n")

# Installed packages
import textwrap

# Local packages
from mezcla import debug
from mezcla import system
from mezcla.tpo_common import format as tpo_format
## TODO3: debug.trace_expr(6, __file__)
from mezcla.validate_arguments_types import (
    FileDescriptorOrPath, StrOrBytesPath
)

# Constants
TL = debug.TL

# Environment options
#
# note:
# - ALLOW_SUBCOMMAND_TRACING should be interepreted in terms of detailed
# tracing. Now, basic tracing is still done unless disable_subcommand_tracing()
# invoked. (This way, the subscript start/end time is still shown by default)
# - SUB_DEBUG_LEVEL added to make sub-script trace level explicit
DEFAULT_SUB_DEBUG_LEVEL = int(min(debug.TL.USUAL, debug.get_level()))
SUB_DEBUG_LEVEL = system.getenv_int(
    "SUB_DEBUG_LEVEL", DEFAULT_SUB_DEBUG_LEVEL,
    description="Tracing level for sub-command scripts invoked")
default_subtrace_level = SUB_DEBUG_LEVEL
ALLOW_SUBCOMMAND_TRACING = system.getenv_boolean(
    "ALLOW_SUBCOMMAND_TRACING",
    (SUB_DEBUG_LEVEL > DEFAULT_SUB_DEBUG_LEVEL),
    description="Whether sub-commands have tracing above TL.USUAL")
if ALLOW_SUBCOMMAND_TRACING:
    # TODO: work out intuitive default if both SUB_DEBUG_LEVEL and ALLOW_SUBCOMMAND_TRACING specified
    default_subtrace_level = max(debug.get_level(), SUB_DEBUG_LEVEL)

INDENT = system.getenv_text(
    "INDENT_TEXT", "    ",
    description="Default indentation")
# 
# note: See main.py for similar support as part of Main scipt class
FILE_BASE = system.getenv_text(
    "FILE_BASE", "_temp",
    description="Basename for output files including dir")
TEMP_PREFIX = system.getenv_text(
    "TEMP_PREFIX", FILE_BASE + "-",
    description="Prefix to use for temp files")
TEMP_SUFFIX = system.getenv_text(
    "TEMP_SUFFIX", "_",
    description="Suffix to use for temp files")
KEEP_TEMP = system.getenv_bool(
    "KEEP_TEMP", debug.detailed_debugging(),
    desc="Keep temporary files")
TEMP_BASE = system.getenv_value(
    "TEMP_BASE", None,
    description="Debugging override for temporary file basename")
USE_TEMP_BASE_DIR_DEFAULT = bool(
    TEMP_BASE and (system.is_directory(TEMP_BASE) or TEMP_BASE.endswith("/")))
USE_TEMP_BASE_DIR = system.getenv_bool(
    "USE_TEMP_BASE_DIR", USE_TEMP_BASE_DIR_DEFAULT,
    description="Whether TEMP_BASE should be a dir instead of prefix")
DISABLE_RECURSIVE_DELETE = system.getenv_value(
    "DISABLE_RECURSIVE_DELETE", debug.detailed_debugging(),
    description="Disable potentially dangerous rm -r style or rmtree recursive deletions")
PRESERVE_TEMP_FILE = None
HOME_DIR = system.getenv_text(
    "HOME", Path.home(),
    description="User home directory")

# Globals
# note:
# - see init() for main initialization;
#   these are placeholders until module initialized below (see init)
# - os.path.join used to likewise avoid chicken-n-egg problems with init
# - TEMP_FILE is normally None to indicate use of random temp file name
# - TEMP_LOG_FILE and TEMP_SCRIPT_FILE are used in run, issue, etc.
# TODO3: make GLOBAL_TEMP_FILE, etc. lowercase
TMP = system.getenv_text(
    "TMP", "/tmp",
    description="Temporary directory")
PID = system.get_process_id()
TEMP_FILE = None
PID_BASENAME = f"temp-{PID}"
GLOBAL_TEMP_FILE = os.path.join(TMP, PID_BASENAME)
TEMP_LOG_FILE = os.path.join(TMP, f"{GLOBAL_TEMP_FILE}.log")
TEMP_SCRIPT_FILE = os.path.join(TMP, f"{GLOBAL_TEMP_FILE}.script")
initialized = False

#------------------------------------------------------------------------

def get_temp_file(delete: Optional[bool] = None) -> str:
    """Return name of unique temporary file, optionally with DELETE"""
    # Note: delete defaults to False if detailed debugging
    # TODO: allow for overriding other options to NamedTemporaryFile
    if ((delete is None) and debug.detailed_debugging()):
        delete = False
    ## BAD: temp_file_name = (TEMP_FILE or tempfile.NamedTemporaryFile(**NTF_ARGS).name)
    NTF_ARGS = {'prefix': TEMP_PREFIX,
                'delete': not KEEP_TEMP,
                'suffix': TEMP_SUFFIX}
    temp_file_name = TEMP_FILE
    if not temp_file_name:
        # note: uses context so not deleted right away if delete=True
        # TODO2: fix this
        with tempfile.NamedTemporaryFile(**NTF_ARGS) as temp_file_obj:
            temp_file_name = temp_file_obj.name
        # HACK: clear the file
        if not KEEP_TEMP:
            system.write_file(temp_file_name, "")
    ## TODO2: drop ... or ""
    temp_file_name = temp_file_name or ""
    debug.assertion(not delete, "Support for delete not implemented")
    debug.trace_fmtd(5, "gh.get_temp_file() => {r!r}", r=temp_file_name)
    return temp_file_name


def get_temp_dir(delete=None) -> str:
    """Gets temporary file to use as a directory
    note: Optionally DELETEs directory afterwards
    """
    ## TODO3: make option to bypass creation
    temp_dir_path = get_temp_file(delete=delete)
    # note: removes non-dir file if exists
    full_mkdir(temp_dir_path, force=True)
    return temp_dir_path


def create_temp_file(contents: Any, binary: bool = False) -> str:
    """Create temporary file with CONTENTS and return full path"""
    temp_filename = get_temp_file()
    system.write_file(temp_filename, contents, binary=binary)
    debug.trace(6, f"create_temp_file({contents!r}) => {temp_filename}")
    return temp_filename


def basename(filename: str, extension: Optional[str] = None) -> str:
    """Remove directory from FILENAME along with optional EXTENSION, as with Unix basename command. Note: the period in the extension must be explicitly supplied (e.g., '.data' not 'data')"""
    # EX: basename("fubar.py", ".py") => "fubar"
    # EX: basename("fubar.py", "py") => "fubar."
    # EX: basename("/tmp/solr-4888.log", ".log") => "solr-4888"
    base = os.path.basename(filename) or ""
    if extension is not None:
        pos = base.find(extension)
        if pos > -1:
            base = base[:pos]
    debug.trace(5, f"basename({filename!r}, {extension}) => {base}")
    return base


def remove_extension(filename: str, extension: str) -> str:
    """Returns FILENAME without EXTENSION. Note: similar to basename() but retaining directory portion."""
    # EX: remove_extension("/tmp/solr-4888.log", ".log") => "/tmp/solr-4888"
    # EX: remove_extension("/tmp/fubar.py", ".py") => "/tmp/fubar"
    # EX: remove_extension("/tmp/fubar.py", "py") => "/tmp/fubar."
    # NOTE: Unlike os.path.splitext, only the specific extension is removed (not whichever extension used).
    pos = filename.find(extension)
    base = filename[:pos] if (pos > -1) else filename
    debug.trace(5, f"remove_extension({filename!r}, {extension}) => {base!r}")
    return base


## TODO: def replace_extension(filename, old_extension, old_extension):
##           ...


def dir_path(filename: str, explicit: bool = False) -> str:
    """Wrapper around os.path.dirname over FILENAME
    Note: With EXPLICIT, returns . instead of "" (e.g., if filename in current directory)
    """
    # TODO: return . for filename without directory (not "")
    # EX: dir_path("/tmp/solr-4888.log") => "/tmp"
    # EX: dir_path("README.md") => ""
    path = os.path.dirname(filename)
    if (not path and explicit):
        path = "."
    debug.trace(5, f"dir_path({filename!r}, [explicit={explicit}]) => {path}")
    # TODO: add realpath (i.e., canonical path)
    if not explicit:
        base = basename(filename)
        debug.assertion(form_path(path, base) == filename)
    return path


def dirname(file_path: str) -> str:
    """"Returns directory component of FILE_PATH as with Unix dirname
    Note: Unlike dir_path, this always returns explicit directory
    """
    # EX: dirname("/tmp/solr-4888.log") => "/tmp"
    # EX: dirname("README.md") => "."
    return dir_path(file_path, explicit=True)


def file_exists(filename: FileDescriptorOrPath) -> bool:
    """Returns indication that FILENAME exists"""
    ok = os.path.exists(filename)
    debug.trace(7, f"file_exists({filename!r}) => {ok}")
    return ok


def non_empty_file(filename: FileDescriptorOrPath) -> bool:
    """Whether FILENAME exists and is non-empty"""
    size = (os.path.getsize(filename) if os.path.exists(filename) else -1)
    non_empty = (size > 0)
    debug.trace(5, f"non_empty_file({filename!r}) => {non_empty}; (filesize={size})")
    return non_empty


def resolve_path(
        filename: str,
        base_dir: Optional[str] = None,
        heuristic: bool = False
    ) -> str:
    """Resolves path for FILENAME or path, relative to BASE_DIR if not in current directory. 
    Note:
    - This uses the script directory for the calling module if BASE_DIR not specified
      (e.g., as if os.path.dirname(__file__) passed).
    - If HEURISTIC, then also checks nearby directories such as parent for BASE_DIR, which
      is useful for resolving resources for tests, which normally run of module dir 
      (e.g., mezcla for mezcla/tests/test_template.py): see test_heuristic_resolve_path.
    - HEURISTIC also uses find.
    """
    ## TODO4: rename filename to sub_path for clarity
    debug.trace(5, f"in resolve_path({filename!r})")
    debug.trace_expr(6,  base_dir, heuristic)
    # TODO: give preference to script directory over current directory
    path = filename
    if not os.path.exists(path):
        # Determine directly for calling module
        if not base_dir:
            frame = None
            try:
                frame = inspect.currentframe()
                assert isinstance(frame, FrameType)
                frame = frame.f_back
                assert isinstance(frame, FrameType)
                calling_filename = frame.f_globals['__file__']
                base_dir = os.path.dirname(calling_filename)
                debug.trace_expr(4, calling_filename, base_dir)
            except (AssertionError, AttributeError, KeyError):
                base_dir = ""
                debug.trace(5, "Error: Exception during resolve_path: " + str(sys.exc_info()))
            finally:
                if frame:
                    del frame
        if not isinstance(base_dir, str):
            base_dir = str(base_dir)
        
        # Check calling directory (TODO2: add more check dirs such as children)
        dirs_to_check = [base_dir]
        if heuristic:
            dirs_to_check += [form_path(base_dir, ".."), form_path(".", "..")]
        for check_dir in dirs_to_check:
            check_path = os.path.join(check_dir, path)
            if os.path.exists(check_path):
                path = check_path
                break
    # Fall back to using find command
    if (not os.path.exists(path)) and heuristic:
        debug.trace(4, "FYI: resolve_path falling back to find")
        debug.assertion(" " not in path)
        debug.assertion(base_dir)
        path = run(f"find {base_dir or '.'} -name '{path}'")
            
    debug.trace_fmtd(4, "resolve_path({f}) => {p}", f=filename, p=path)
    return path


def form_path(*filenames: str, create: bool = False) -> str:
    """Wrapper around os.path.join over FILENAMEs (with tracing)
    Note: includes sanity check about absolute filenames except for first
    If CREATE, then the directory for the path is created if needed
    Warning: This might be deprecated: use system.form_path instead.
    """
    ## TODO3: return system.form_path(*filenames)
    debug.assertion(not any(f.startswith(system.path_separator()) for f in filenames[1:]))
    if create:
        path_dir = os.path.join(*filenames[:-1])
        if not system.file_exists(path_dir):
            full_mkdir(path_dir)

    path = os.path.join(*filenames)
    debug.trace_fmtd(6, "form_path{f} => {p}", f=tuple(filenames), p=path)
    return path


def is_directory(path: FileDescriptorOrPath) -> bool:
    """Determines whether PATH represents a directory"""
    is_dir = os.path.isdir(path)
    debug.trace_fmtd(6, "is_dir({p}) => {r}", p=path, r=is_dir)
    return is_dir


## TODO2: add decorator for flagging obsolete functions
##   def obsolete():
##      """Flag fucntion as obsolete in docstring and issue warning if called"""
##      warning = f"Warning {func} obsolete use version in system.py instead"
##      func.docstring += warning
##      func.body = f'debug.trace(3, "{warning}")' + func.body


def create_directory(path: StrOrBytesPath) -> None:
    """Wrapper around os.mkdir over PATH (with tracing)
    Warning: obsolete
    """
    debug.trace(3, "Warning: create_directory obsolete use version in system.py instead")
    system.create_directory(path)
    return


def full_mkdir(path: FileDescriptorOrPath, force: bool = False) -> FileDescriptorOrPath:
    """Issues mkdir to ensure path directory, including parents (assuming Linux like shell)
    Note:
    - When FORCE true, an existing non-directory is removed first.
    - Otherwise, doesn't handle case when file exists but is not a directory.
    - Returns path, which is useful for temporary sub-directory creation:
    -   gh.full_mkdir(gh.form_path(gh.get_temp_dir(), 'my_temp_subdir'))
    """
    debug.trace(6, f"full_mkdir({path!r})")
    debug.assertion(os.name == "posix")
    if force and system.file_exists(path) and not system.is_directory(path):
        delete_file(path)
    if not system.file_exists(path):
        os.makedirs(path, exist_ok=True)
    debug.assertion(is_directory(path))
    return path


def real_path(path: FileDescriptorOrPath) -> str:
    """Return resolved absolute pathname for PATH, as with Linux realpath command
    Note: Use version in system instead"""
    # EX: re.search("vmlinuz.*\d.\d", real_path("/vmlinuz"))
    ## TODO: result = system.real_path(path)
    ## TODO3: check outher modules for over-applicaiton of !r-formatting
    ## BAD: result = run(f'realpath "{path!r}"')
    result = run(f'realpath "{path}"')
    debug.trace(6, "Warning: obsolete: use system.real_path instead")
    debug.trace(7, f"real_path({path!r}) => {result}")
    return result


def indent(text: str, indentation: Optional[str] = None, max_width: int = 512) -> str:
    """Indent TEXT with INDENTATION at beginning of each line, returning string ending in a newline unless empty and with resulting lines longer than max_width characters wrapped. Text is treated as a single paragraph."""
    if indentation is None:
        indentation = INDENT
    # Note: an empty text is returned without trailing newline
    tw = textwrap.TextWrapper(width=max_width, initial_indent=indentation, subsequent_indent=indentation)
    wrapped_text = "\n".join(tw.wrap(text))
    if wrapped_text and text.endswith("\n"):
        wrapped_text += "\n"
    return wrapped_text


def indent_lines(text: str, indentation: Optional[str] = None, max_width: int = 512) -> str:
    """Like indent, except that each line is indented separately. That is, the text is not treated as a single paragraph."""
    # Sample usage: print("log contents: {{\n{log}\n}}".format(log=indent_lines(lines)))
    # TODO: add support to simplify above idiom (e.g., indent_lines_bracketed); rename to avoid possible confusion that input is array (as wih write_lines)
    if indentation is None:
        indentation = INDENT
    result = ""
    for line in text.splitlines():
        indented_line = indent(line + "\n", indentation, max_width)
        if not indented_line:
            indented_line = "\n"
        result += indented_line
    return result


MAX_ELIDED_TEXT_LEN = system.getenv_integer("MAX_ELIDED_TEXT_LEN", 128)
#
def elide(value: Optional[Any], max_len: Optional[int] = None) -> str:
    """Returns VALUE converted to text and elided to at most MAX_LEN characters (with '...' used to indicate remainder). 
    Note: intended for tracing long strings."""
    # EX: elide("=" * 80, max_len=8) => "========..."
    # EX: elide(None) => ""
    # NOTE: Make sure compatible with debug.format_value (TODO3: add equivalent to strict argument)
    # TODO2: add support for eliding at word-boundaries
    debug.trace(8, "elide(_, _)")
    text = value
    if text is None:
        text = ""
    if (not isinstance(text, str)):
        text = str(text)
    if max_len is None:
        max_len = MAX_ELIDED_TEXT_LEN
    result = text
    if (result and (len(result) > max_len)):
        result = result[:max_len] + "..."
    debug.trace(9, "elide({%s}, [{%s}]) => {%s}" % (text, max_len, result))
    return result
#
# EX: elide(None, 10) => ''

def elide_values(values: List[Any], **kwargs) -> List[str]:
    """List version of elide [q.v.]"""
    # EX: elide_values(["1", "22", "333"], max_len=2) => ["1", "22", "33..."]
    debug.trace(7, "elide_values(_, _)")
    return list(map(lambda v: elide(str(v), **kwargs),
                    values))


def disable_subcommand_tracing() -> None:
    """Disables tracing in scripts invoked via run().
    Note: Invoked in unittest_wrapper.py"""
    debug.trace(7, "disable_subcommand_tracing()")
    # Note this works by having run() temporarily setting DEBUG_LEVEL to 0."""
    global default_subtrace_level
    default_subtrace_level = 0


def run(
        command: str,
        trace_level: debug.IntOrTraceLevel = 4,
        subtrace_level: Optional[debug.IntOrTraceLevel] = None,
        just_issue: Optional[bool] = None,
        output: bool = False,
        **namespace
    ) -> str:
    """Invokes COMMAND via system shell (e.g., os.system), using TRACE_LEVEL for debugging output, returning result. The command can use format-style templates, resolved from caller's namespace. The optional SUBTRACE_LEVEL sets tracing for invoked commands (default is same as TRACE_LEVEL); this works around problem with stderr not being separated, which can be a problem when tracing unit tests.
   Notes:
   - The result includes stderr, so direct if not desired (see issue):
         run("ls /tmp/fubar 2> /dev/null")
   - This is only intended for running simple commands. It would be better to create a subprocess for any complex interactions.
   - This function doesn't work fully under Win32. Tabs are not preserved, so redirect stdout to a file if needed.
   - If TEMP_FILE or TEMP_BASE defined, these are modified to be unique to avoid conflicts across processeses.
    - If OUTPUT, the result will be printed.
   """
    # TODO: add automatic log file support as in run_script from unittest_wrapper.py
    # TODO: make sure no template markers left in command text (e.g., "tar cvfz {tar_file}")
    # EX: "root" in run("ls /")
    # Note: Script tracing controlled DEBUG_LEVEL environment variable.
    debug.assertion(isinstance(trace_level, int))
    debug.trace(trace_level + 2, f"run({command}, tl={trace_level}, sub_tr={subtrace_level}, iss={just_issue}, out={output})", skip_sanity_checks=True)
    global default_subtrace_level
    # Keep track of current debug level settings for later restoration
    debug_level_env = os.getenv("DEBUG_LEVEL")
    sub_debug_level_env = os.getenv("SUB_DEBUG_LEVEL")
    if subtrace_level is None:
        subtrace_level = default_subtrace_level
    if subtrace_level != system.to_int(debug_level_env, ignore=True):
        system.setenv("DEBUG_LEVEL", str(subtrace_level))
    if subtrace_level != system.to_int(sub_debug_level_env, ignore=True):
        # note: for run/issue called within scripts
        system.setenv("SUB_DEBUG_LEVEL", str(subtrace_level))
    in_just_issue = just_issue
    if just_issue is None:
        just_issue = False
    save_temp_base = TEMP_BASE
    if TEMP_BASE:
        # note: makes sure subprocess TEMP_BASE is dir if main one is
        if system.is_directory(TEMP_BASE) or TEMP_BASE.endswith("/"):
            system.create_directory(TEMP_BASE)
            new_TEMP_BASE = form_path(TEMP_BASE, "_subprocess_", create=True)
            system.setenv("TEMP_BASE", new_TEMP_BASE)
            ## TEMP
            system.create_directory(new_TEMP_BASE)
        else:
            system.setenv("TEMP_BASE", TEMP_BASE + "_subprocess_")
    save_temp_file = TEMP_FILE
    if TEMP_FILE and (PRESERVE_TEMP_FILE is not True):
        new_TEMP_FILE = TEMP_FILE + "_subprocess_"
        debug.trace_expr(5, PRESERVE_TEMP_FILE)
        debug.trace(5, f"Setting TEMP_FILE to {new_TEMP_FILE}")
        system.setenv("TEMP_FILE", new_TEMP_FILE)
    # Expand the command template if brace-style variable reference encountered
    # NOTE: un-pythonic warnings issued by format so this should not affect anything
    # TODO: make this optional
    command_line = command
    if (re.search(r"{\S+}", command) or namespace):
        if not namespace:
            ## TODO3: weed out gh.run calls with empty kwargs
            debug.trace(4, "Warning: deprecated interpolation used for gh.run")
        command_line = tpo_format(command_line, indirect_caller=True, ignore_exception=False, **namespace)
    else:
        # TODO2: and sanity check for unresolved f-string-like template as with debug.trace
        pass
    debug.trace(trace_level, "issuing: %s" % command_line)
    # Run the command
    # TODO: check for errors (e.g., "sh: filter_file.py: not found"); make wait explicit
    in_background = command.strip().endswith("&")
    foreground_wait = not in_background
    debug.trace_expr(5, in_background, in_just_issue)
    debug.assertion(not (in_background and (in_just_issue is False)))
    # Note: Unix supports the '>|' pipe operator (i.e., output with overwrite); but,
    # it is not supported under Windows. To avoid unexpected porting issues, clients
    # should replace 'run("... >| f")' usages with 'delete_file(f); run(...)'.
    # note: TestWrapper.setUp handles the deletion automatically
    debug.assertion(">|" not in command_line)
    result = None
    ## TODO: if (just_issue or not foreground_wait): ... else: ...
    wait_for_command = (foreground_wait and not just_issue)
    debug.trace_expr(5, foreground_wait, just_issue, wait_for_command)
    ## TODO3: clarify what output is when stdout redirected (e.g., for issue in support of unittest_wrapper.run_script
    result = getoutput(command_line) if wait_for_command else str(os.system(command_line))
    if output:
        print(result)
    # Restore debug level setting in environment
    system.setenv("DEBUG_LEVEL", debug_level_env or "")
    system.setenv("SUB_DEBUG_LEVEL", sub_debug_level_env or "")
    system.setenv("TEMP_BASE", save_temp_base or "")
    if save_temp_file and (PRESERVE_TEMP_FILE is not True):
        debug.trace(5, f"Resetting TEMP_FILE to {save_temp_file}")
        system.setenv("TEMP_FILE", save_temp_file)
    debug.trace_fmt((trace_level + 1), "run(_) => {{\n{r}\n}}", r=indent_lines(result))
    return result


def run_via_bash(
        command: str,
        trace_level: debug.IntOrTraceLevel = 4,
        subtrace_level: Optional[debug.IntOrTraceLevel] = None,
        init_file: Optional[bool] = None,
        enable_aliases: bool = False,
        **namespace
    ) -> str:
    """Version of run that runs COMMAND with aliases defined
    Notes:
    - This can be slow due to alias definition overhead
    - INIT_FILE is file to source before running the command
    - TRACE_LEVEL and SUBTRACE_LEVEL control tracing for COMMAND and any subcommands, respectively
    - Used in bash to python translation; see
         https://github.com/tomasohara/shell-scripts/blob/main/bash2python.py
    """
    debug.trace(trace_level, "issuing: %s" % command)
    commands_to_run = ""
    if enable_aliases:
        commands_to_run += "shopt -s expand_aliases\n"
    if init_file:
        commands_to_run += system.read_file(init_file) + "\n"
    commands_to_run += command
    system.write_file(TEMP_SCRIPT_FILE, commands_to_run)
    
    command_line = f"bash -f {TEMP_SCRIPT_FILE}"
    return run(command_line, trace_level=(trace_level + 1), subtrace_level=subtrace_level, just_issue=False, **namespace)


def issue(
        command: str,
        trace_level: debug.IntOrTraceLevel = 4,
        subtrace_level: Optional[debug.IntOrTraceLevel] = None,
        log_file: Optional[str] = None,
        **namespace
    ) -> None:
    """Wrapper around run() for when output is not being saved (i.e., just issues command). 
    Note:
    - Nothing is returned.
    - Traces stdout when debugging at quite-detailed level (6).
    - Captures stderr unless redirected and traces at error level (1)."""
    # EX: issue("ls /") => None
    # EX: issue("xeyes &")
    debug.trace_fmt(
        (trace_level + 1), "issue({c}, [trace_level={tl}], [sub_level={sl}], [ns={n}])",
        c=command, tl=trace_level, sl=subtrace_level, n=namespace)
    # Add stderr redirect to temporary log file, unless redirection already present
    log_file = None
    has_stderr_redir = ("2>" in command) or ("2|&1" in command)
    if (not log_file) and debug.debugging() and (not has_stderr_redir):
        ## TODO: use a different suffix each time to aid in debugging
        log_file = TEMP_LOG_FILE
    if log_file:
        debug.assertion(not has_stderr_redir)
        delete_existing_file(log_file)
        command += " 2> " + log_file
    # Run the command and trace output
    command_line = command
    if re.search("{.*}", command_line):
        if not namespace:
            ## TODO3: weed out gh.issue calls with empty kwargs
            debug.trace(4, "Warning: deprecated interpolation used for gh.issue")
        command_line = tpo_format(command_line, indirect_caller=True, ignore_exception=False, **namespace)
    output = run(command_line, trace_level, subtrace_level, just_issue=True)
    debug.trace((2 + trace_level), "stdout from command: {\n%s\n}\n" % indent(output))
    # Trace out any standard error output and remove temporary log file (unless debugging)
    if log_file:
        if debug.debugging() and non_empty_file(log_file):
            stderr_output = indent(read_file(log_file))
            debug.trace(1, "stderr output from command: {\n%s\n}\n" % indent(stderr_output))
        if not debug.detailed_debugging():
            ## TODO4: add option for deletion
            delete_file(log_file)
    return


def get_hex_dump(text: Any, break_newlines: bool = False) -> str:
    """Get hex dump for TEXT, optionally BREAKing lines on NEWLINES"""
    # TODO: implement entirely within Pyton (e.g., via binascii.hexlify)
    # EX: get_hex_dump("Tomás") => \
    #   "00000000  54 6F 6D C3 A1 73       -                          Tom..s"
    debug.trace_fmt(6, "get_hex_dump({t}, {bn})", t=text, bn=break_newlines)
    in_file = get_temp_file() + ".in.list"
    out_file = get_temp_file() + ".out.list"
    system.write_file(in_file, text, skip_newline=True)
    run("perl -Ss hexview.perl {i} > {o}", i=in_file, o=out_file)
    result = read_file(out_file).rstrip("\n")
    debug.trace(7, f"get_hex_dump() => {result}")
    return result


def extract_matches(
        pattern: str,
        lines: List[str],
        fields: int = 1,
        multiple: bool = False,
        re_flags: int = 0,
        para_mode: Optional[bool] = False
    ) -> List[str]:
    """Checks for PATTERN matches in LINES of text returning list of tuples with replacement groups.
    Notes: The number of FIELDS can be greater than 1.
    Optionally allows for MULTIPLE matches within a line.
    The lines are concatenated if DOTALL
    Matching optionally uses perl-style PARA_MODE
    """
    ## TODO: If unspecified, the regex flags default to DOTALL.
    # ex: extract_matches(r"^(\S+) \S+", ["John D.", "Jane D.", "Plato"]) => ["John", "Jane"]
    # Note: modelled after extract_matches.perl
    # TODO: make multiple the default
    debug.trace(6, "extract_matches(%s, _, [fld=%s], [m=%s], [flg=%s], [para=%s])" % (pattern, fields, multiple, re_flags, para_mode))
    ## TODO
    ## if re_flags is None:
    ##     re_flags = re.DOTALL
    debug.trace_values(6, lines, "lines")
    debug.assertion(isinstance(lines, list))
    if pattern.find("(") == -1:
        pattern = "(" + pattern + ")"
    if (re_flags and (re_flags & re.DOTALL)):
        lines = [ "\n".join(lines) + "\n" ]
        debug.trace_expr(6, lines)
    if para_mode:
        lines = re.split(r"\n\s*\n", ("\n".join(lines)))
    matches = []
    for i, line in enumerate(lines):
        while line:
            debug.trace(6, f"L{i}: {line}")
            try:
                # Extract match field(s)
                debug.trace_expr(7, pattern, line, re_flags, fields)
                match = re.search(pattern, line, flags=re_flags)
                if not match:
                    break
                result = match.group(1) if (fields == 1) else [match.group(i + 1) for i in range(fields)]
                matches.append(result)
                if not multiple:
                    break

                # Revise line
                debug.assertion(match.end() > 0)
                new_line = line[match.end():]
                if (new_line == line):
                    break
                line = new_line
            except (re.error, IndexError):
                debug.trace(2, "Warning: Exception in pattern matching: %s" % str(sys.exc_info()))
                line = ""
    debug.trace(7, "extract_matches() => %s" % (matches))
    double_indent = INDENT + INDENT
    debug.trace_fmtd(8, "{ind}input lines: {{\n{res}\n{ind}}}",
                 ind=INDENT, res=indent_lines("\n".join(lines), double_indent))
    return matches


def extract_match(
        pattern: str,
        lines: List[str],
        fields: int = 1,
        multiple: bool = False,
        re_flags: int = 0,
        para_mode: Optional[bool] = None
    ):
    """Extracts first match of PATTERN in LINES for FIELDS"""
    matches = extract_matches(pattern, lines, fields, multiple, re_flags, para_mode)
    result = (matches[0] if matches else None)
    debug.trace(5, "match: %s" % result)
    return result


def extract_match_from_text(
        pattern: str,
        text: str,
        fields: int = 1,
        multiple: bool = False,
        re_flags: int = 0,
        para_mode: Optional[bool] = None
    ) -> Optional[str]:
    """Wrapper around extract_match for text input"""
    ## TODO: rework to allow for multiple-line matching
    return extract_match(pattern, text.split("\n"), fields, multiple, re_flags, para_mode)


def extract_matches_from_text(
        pattern: str,
        text: str,
        fields: int = 1,
        multiple: Optional[bool] = None,
        re_flags: int = 0,
        para_mode: Optional[bool] = None
    ) -> List[str]:
    """Wrapper around extract_matches for text input
    Note: By default MULTIPLE matches are returned"""
    # EX: extract_matches_from_text(".", "abc") => ["a", "b", "c"]
    # EX: extract_matches_from_text(".", "abc", multiple=False) => ["a"]
    if multiple is None:
        multiple = True
    # TODO: make multiple True by default
    return extract_matches(pattern, text.split("\n"), fields, multiple, re_flags, para_mode)


def extract_pattern(pattern: str, text: str, **kwargs) -> Optional[str]:
    """Yet another wrapper around extract_match for text input"""
    return extract_match(pattern, text.split("\n"), **kwargs)

def count_it(
        pattern: str,
        text: str,
        field: int = 1,
        multiple: Optional[None] = None
    ) -> Dict[str, int]:
    """Counts how often PATTERN's FIELD occurs in TEXT, returning hash.
    Note: By default MULTIPLE matches are tabulated"""
    # EX: dict(count_it("[a-z]", "Panama")) => {"a": 3, "n": 1, "m": 1}
    # EX: count_it("\w+", "My d@wg's fleas have fleas")["fleas"] => 2
    debug.trace(7, f"count_it({pattern}, _, {field}, {multiple}")
    value_counts: Dict[str, int] = defaultdict(int)
    for value in extract_matches_from_text(pattern, text, field, multiple):
        value_counts[value] += 1
    debug.trace_values(6, value_counts, "count_it()")
    return value_counts


def read_lines(
        filename: Optional[FileDescriptorOrPath] = None,
        make_unicode: bool = False
    ) -> List[str]:
    """Returns list of lines from FILENAME without newlines (or other extra whitespace)
    Note:
    - Uses stdin if filename is None. Optionally returned as unicode.
    - make_unicode is deprecated
    - Warning: deprecated function--use system.read_lines instead
    """
    debug.trace(3, "Warning: in deprecated glue_helpers.read_lines: use version in system")
    debug.assertion(not make_unicode)
    if not filename:
        filename = get_temp_file() + ".stdin.list"
        system.write_file(filename, system.read_all_stdin())
    return system.read_lines(filename)


def write_lines(
        filename: FileDescriptorOrPath,
        text_lines: List[str],
        append: bool = False
    ) -> None:
    """Creates FILENAME using TEXT_LINES with newlines added and optionally for APPEND
    Warning: deprecated function--use version in system.py instead"""
    debug.trace(3, "Warning: in deprecated glue_helpers.read_file: use version in system")
    system.write_lines(filename, text_lines, append=append)


def read_file(filename: FileDescriptorOrPath, make_unicode: bool = False) -> str:
    """Returns text from FILENAME (single string), including newline(s).
    Note: optionally returned as unicde.
    Warning: deprecated function--use system.read_file instead
    """
    debug.trace(3, "Warning: in deprecated glue_helpers.read_file: use version in system")
    debug.assertion(not make_unicode)
    return system.read_file(filename)


def write_file(filename: FileDescriptorOrPath, text: str, append: bool=False) -> None:
    """Writes FILENAME using contents in TEXT, adding trailing newline and optionally for APPEND
    Warning: deprecated function--use system.write_file instead
    """
    debug.trace(3, "Warning: in deprecated glue_helpers.write_file: use version in system")
    system.write_file(filename, text, append=append)


def copy_file(source: str, target: str) -> None:
    """Copy SOURCE file to TARGET file (or directory)"""
    # Note: meta data is not copied (e.g., access control lists)); see
    #    https://docs.python.org/2/library/shutil.html
    # TODO: have option to skip if non-dir target exists
    debug.trace(5, f"copy_file({source}, {target}")
    debug.assertion(non_empty_file(source))
    shutil.copy(source, target)
    if system.is_regular_file(target):
        target_file = target
    else:
        base = basename(source)
        assert isinstance(base, str)
        target_file = form_path(target, base)
    ## TODO: debug.assertion(file_size(source) == file_size(target_file))
    debug.assertion(non_empty_file(target_file))
    return


def non_empty_directory(path):
    """Whether PATH exists and is not empty"""
    size = len(get_directory_listing(path)) if is_directory(path) else -1
    non_empty = size > 0
    debug.trace_fmt(5, f"non_empty_directory({path}) => {non_empty}; (#files={size})")
    return non_empty


def copy_directory(source, dest):
    """copy SOURCE dir to DEST dir
    Note: The DEST directory must not exist beforehand
    """
    ## TODO4: add option to overwrite files (e.g., via copytree's dirs_exist_ok)
    # Note: meta data is not copied (e.g., access control lists)); see
    #    https://docs.python.org/3/library/shutil.html
    debug.trace_fmt(5, f"copy_directory({source}, {dest})")

    ## OLD: debug.assertion(non_empty_directory(source))
    dest_path = shutil.copytree(src=source, dst=dest)
    debug.assertion(len(get_directory_listing(source)) == len(get_directory_listing(dest_path)))
    ## OLD: debug.assertion(non_empty_directory(dest_path))


def rename_file(source: StrOrBytesPath, target: str) -> None:
    """Rename SOURCE file as TARGET file"""
    # TODO: have option to skip if target exists
    debug.trace(5, f"rename_file({source}, {target})")
    debug.assertion(non_empty_file(source))
    debug.assertion(source != target)
    os.rename(source, target)
    debug.assertion(non_empty_file(target))
    return


def delete_file(filename: StrOrBytesPath) -> bool:
    """Deletes FILENAME"""
    debug.trace(5, f"delete_file({filename})")
    debug.assertion(os.path.exists(filename))
    ok = False
    try:
        os.remove(filename)
        ok = True
        debug.trace_fmtd(6, "remove{f} => {r}", f=filename, r=ok)
    except OSError:
        ## OLD: debug.trace(5, "Exception during deletion of {filename}: " + str(sys.exc_info()))
        debug.trace_exception(5, f"deletion of {filename}")
    return ok


def delete_existing_file(filename: StrOrBytesPath) -> bool:
    """Deletes FILENAME if it exists and is not a directory or other special file"""
    ok = False
    if file_exists(filename):
        ok = delete_file(filename)
    debug.trace_fmtd(5, "delete_existing_file({f}) => {r}", f=filename, r=ok)
    return ok

def delete_directory(path):
    """Deletes directory at PATH
    Warning: Unless DISABLE_RECURSIVE_DELETE, this removes entire directory tree
    """
    debug.trace_fmt(5, f"delete_directory({path})")
    ok = False
    try:
        if DISABLE_RECURSIVE_DELETE:
            files = get_directory_listing(path)
            debug.trace(4, f"FYI: Only deleting top-level files in {path} to avoid potentially dangerous recursive deletion")
            for file in files:
                delete_file(form_path(path, file))
            ok = None
        else:
            debug.trace(4, f"FYI: Using potentially dangerous rmtree over {path}")
            shutil.rmtree(path)
            ok = None
    except OSError:
        debug.trace_fmt(5, f"Exception during deletion of {path}: {system.get_exception()}")
    return ok

def file_size(filename: FileDescriptorOrPath) -> int:
    """Returns size of FILENAME in bytes (or -1 if not found)"""
    size = -1
    if os.path.exists(filename):
        size = os.path.getsize(filename)
    debug.trace_fmtd(5, "file_size({f}) => {s}", f=filename, s=size)
    return size


def get_matching_files(pattern: str, warn: bool = False) -> List[str]:
    """Get sorted list of files matching PATTERN via shell globbing
    Note: Optionally issues WARNing"""
    # NOTE: Multiple glob specs not allowed in PATTERN
    files = sorted(glob.glob(pattern))
    debug.trace_fmtd(5, "get_matching_files({p}) => {l}",
                     p=pattern, l=files)
    if ((not files) and warn):
        system.print_stderr(f"Warning: no matching files for {pattern}")
    return files


def get_files_matching_specs(patterns: List[str]) -> List[str]:
    """Get list of files matching PATTERNS via shell globbing"""
    files = []
    for spec in patterns:
        files += get_matching_files(spec)
    debug.trace_fmtd(6, "get_files_matching_specs({p}) => {l}",
                     p=patterns, l=files)
    return files


def get_directory_listing(
        dir_name: Union[int, str, bytes],
        make_unicode: bool = False
    ) -> Union[List[str], List[str], List[bytes]]:
    """Returns files in DIR_NAME
    Note: make_unicode is deprecated
    """
    # TODO: Union[List[str], List[bytes]] = []
    ## TODO3: drop make_unicode; implement via system.get_directory_filenames
    all_file_names: Union[List[str], List[str], List[bytes]] = []
    try:
        all_file_names = os.listdir(dir_name)
    except OSError:
        debug.trace_fmtd(4, "Exception during get_directory_listing: {exc}",
                         exc=str(sys.exc_info()))
    if make_unicode:
        debug.trace(4, "Warning: using obsolete get_directory_listing make_unicode option")
    debug.trace_fmtd(5, "get_directory_listing({dir}) => {files}",
                     dir=dir_name, files=all_file_names)
    return all_file_names

#-------------------------------------------------------------------------------
# Extensions to system.py included here due to inclusion of functions 
# defined here.
## TODO3: put in system.py

def getenv_filename(var: str, default: str = "", description: Optional[str] = None) -> str:
    """Returns text filename based on environment variable VAR (or string version of DEFAULT) 
    with optional DESCRIPTION. This includes a sanity check for file being non-empty."""
    # EX: system.setenv("ETC", "/etc"); getenv_filename("ETC") => "/etc"
    # TODO4: explain motivation
    debug.trace_fmtd(6, "getenv_filename({v}, {d}, {desc})",
                 v=var, d=default, desc=description)
    if not description:
        description = ""
    filename = system.getenv_text(var, default, description)
    if filename and not non_empty_file(filename):
        system.print_stderr("Error: filename %s empty or missing for environment option %s" % (filename, var))
    return filename


if __debug__:

    assertion_deprecation_shown = False
    
    def assertion(condition: bool) -> None:
        """Issues warning if CONDITION doesn't hold
        Note: deprecated function--use debug.assertion instead"""
        global assertion_deprecation_shown
        if not assertion_deprecation_shown:
            debug.trace(3, "Warning: glue_helpers.assertion() is deprecated; use version in debug.py")
            assertion_deprecation_shown = True
        if debug.assertion(condition, indirect=True):
            try:
                frame = inspect.currentframe()
                frame = frame.f_back
                filename = frame.f_globals.get("__file__")
                line_num = frame.f_lineno
                debug.trace(TL.WARNING, f"FYI: Assertion failed at {filename}:{line_num}")
            except:
                system.print_exception_info("gh.assertion")
        return

else:

    def assertion(_condition: bool) -> None:
        """Non-debug stub for assertion"""
        return

def init() -> None:
    """Work around for Python quirk
    Note: This is also used for reinitialize temp-file settings such as for unit tests (e.g., TEMP_FILE from TEMP_BASE).
    Warning: The environment is used to reset following globals:
        PRESERVE_TEMP_FILE, TEMP_FILE, TEMP_LOG_FILE, TEMP_SCRIPT_FILE
    """
    global initialized
    # See https://stackoverflow.com/questions/1590608/how-do-i-forward-declare-a-function-to-avoid-nameerrors-for-functions-defined
    debug.trace(5, "glue_helpers.init()")
    ## OLD: temp_filename = f"{PID_BASENAME}.list"
    temp_filename = f"{PID_BASENAME}"
    if USE_TEMP_BASE_DIR and TEMP_BASE:
        ## OLD: full_mkdir(TEMP_BASE)
        pass

    # Re-initialize flag blocking TEMP_FILE init from TEMP_BASE
    global PRESERVE_TEMP_FILE
    PRESERVE_TEMP_FILE = system.getenv_bool(
        "PRESERVE_TEMP_FILE", None, allow_none=True, skip_register=initialized,
        desc="Retain value of TEMP_FILE even if TEMP_BASE set--see run and init below as well as unittest_wrapper.py")

    # note: Normally TEMP_FILE gets overriden when TEMP_BASE set. However,
    # this complicates preserving test-specific test files (see unittest_wrapper.py).
    # Further compications are due to the implicit module loading due to __init__.py.
    # See tests/test_unittest_wrapper.py for some diagnosis tips.
    temp_file_default = None
    if TEMP_BASE and not PRESERVE_TEMP_FILE:
        temp_file_default = (form_path(TEMP_BASE, temp_filename) if USE_TEMP_BASE_DIR else f"{TEMP_BASE}-{temp_filename}")
        debug.trace(4, f"FYI: Inferred TEMP_FILE default: {temp_file_default!r}")
    debug.trace_expr(5, system.getenv("TEMP_FILE"))
    global TEMP_FILE
    TEMP_FILE = system.getenv_value(
        "TEMP_FILE", temp_file_default, skip_register=initialized,
        description="Debugging override for temporary filename: avoid if possible")
    debug.trace_expr(5, system.getenv("TEMP_FILE"))
    #
    global TEMP_LOG_FILE
    TEMP_LOG_FILE = system.getenv_text(
        "TEMP_LOG_FILE", get_temp_file() + "-log", skip_register=initialized,
        description="Log file for stderr such as for issue function")
    global TEMP_SCRIPT_FILE
    TEMP_SCRIPT_FILE = system.getenv_text(
        "TEMP_SCRIPT_FILE", get_temp_file() + "-script", skip_register=initialized,
        description="File for command invocation")
    initialized = True
#
init()

def main() -> None:
    """Entry point"""
    # Uses dynamic import to avoid circularity
    from mezcla.main import Main        # pylint: disable=import-outside-toplevel
    
    # Note: Uses main-based arg parsing for sake of show environment options
    #   ./glue_helpers.py --help --verbose
    debug.trace(TL.USUAL, f"main(): script={system.real_path(__file__)}")

    # Parse command line options, show usage if --help given
    main_app = Main(description=__doc__.format(script=basename(__file__)))
    debug.assertion(main_app.parsed_args)
    return
    
#------------------------------------------------------------------------

# Warn if invoked standalone
#
if __name__ == '__main__':
    debug.trace_current_context(level=TL.QUITE_VERBOSE)
    system.print_stderr(f"Warning: {__file__} is not intended to be run standalone\n")
    main()



#! /usr/bin/env python3
# TODO: # -*- coding: utf-8 -*-
#
# Based on following:
#   TODO: url
#

"""
TODO: what module does (brief)

Sample usage:
   echo $'TODO:task1\\nDONE:task2' | {script} --TODO-arg --
"""

# Standard modules
## TODO: import json

# Intalled module
# TODO: import numpy as np

# Local modules
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import Main
## TODO: from mezcla.my_regex import my_re
from mezcla import system
## TODO2: streamline imports by exposing common functions, etc. in mezcla

# Constants
TL = debug.TL
## TODO: TODO_BOOL_OPT = "todo-bool-option"
## TODO: TODO_TEXT_OPT = "todo-text-option"

## TODO:
## # Environment options
## # Notes:
## # - These are just intended for internal options, not for end users.
## # - They also allow for enabling options in one place rather than four
## #   when using main.Main (e.g., [Main member] initialization, run-time
## #   value, and argument spec., along with string constant definition).
## #
## ENABLE_FUBAR = system.getenv_bool("ENABLE_FUBAR", False,
##                                   description="Enable fouled up beyond all recognition processing")

#-------------------------------------------------------------------------------

def main():
    """Entry point"""
    debug.trace(TL.USUAL, f"main(): script={system.real_path(__file__)}")

    # Parse command line options, show usage if --help given
    # TODO: manual_input=True; short_options=True
    # Note: Uses Main without subclassing, so some methods are stubs (e.g., run_main_step).
    main_app = Main(
        description=__doc__.format(script=gh.basename(__file__)),
        ## TODO: boolean_options=[(TODO_BOOL_OPT, "TODO desc1")],
        ## TODO: text_options=[(TODO_TEXT_OPT, "TODO desc2")],
    )
    debug.assertion(main_app.parsed_args)
    ## TODO_opt1 = main_app.get_parsed_option(TODO_BOOL_OPT)

    ## TODO:
    system.print_error("Error: Implement me!")
    ## ALT TODO:
    ## for line in main_app.read_entire_input().splitlines():
    ##     print(TODO_fn(line))
    return

#-------------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context(level=TL.QUITE_VERBOSE)
    main()
